#!/usr/bin/env python3

"""
Dialog for managing local caches
"""

from enum import IntEnum
import os
import platform
import shutil
import stat
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2
from cruiz.commands.logdetails import LogDetails

from cruiz.interop.pod import ConanHook
from cruiz.settings.managers.namedlocalcache import (
    NamedLocalCacheSettings,
    NamedLocalCacheSettingsReader,
    NamedLocalCacheSettingsWriter,
    AllNamedLocalCacheSettingsReader,
    NamedLocalCacheDeleter,
    _EnvChangeManagement,
)
from cruiz.settings.managers.recentconanremotes import (
    RecentConanRemotesSettingsReader,
    RecentConanRemotesSettingsWriter,
)
from cruiz.widgets.util import BlockSignals
from cruiz.commands.context import ConanContext, ConanConfigBoolean

from cruiz.revealonfilesystem import reveal_on_filesystem
from cruiz.constants import DEFAULT_CACHE_NAME

if PYSIDE2:
    from cruiz.pyside2.local_cache_manage import Ui_ManageLocalCaches

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.local_cache_manage import Ui_ManageLocalCaches

    QAction = QtGui.QAction

from .widgets import (
    NewLocalCacheWizard,
    InstallConfigDialog,
    MoveLocalCacheDialog,
    AddRemoteDialog,
    RemoveLocksDialog,
    RemoveAllPackagesDialog,
    AddExtraProfileDirectoryDialog,
    AddEnvironmentDialog,
    RemoveEnvironmentDialog,
    RunConanCommandDialog,
)


class ManageLocalCachesDialog(QtWidgets.QDialog):
    """
    New Manage Local Caches dialog.
    """

    _modified = QtCore.Signal()
    cache_changed = QtCore.Signal(str)

    class _HooksTableColumnIndex(IntEnum):
        """
        Column indices of the hooks table
        """

        ENABLED = 0
        PATH = 1

    def __init__(
        self,
        parent: QtWidgets.QWidget,
        cache_name_to_open: typing.Optional[str],
    ) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        self._ui = Ui_ManageLocalCaches()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._ui.new_local_cache_button.clicked.connect(self._new_cache)
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).setEnabled(False)
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).clicked.connect(
            self._save_modifications
        )
        # locations
        dir_icon = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
        self._ui.conan_user_home.setReadOnly(True)
        open_user_home_action = QAction(dir_icon, "", self)
        open_user_home_action.triggered.connect(self._open_conan_user_home)
        self._ui.conan_user_home.addAction(
            open_user_home_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.conan_user_home_short.setReadOnly(True)
        open_short_user_home_action = QAction(dir_icon, "", self)
        open_short_user_home_action.triggered.connect(self._open_conan_user_home_short)
        self._ui.conan_user_home_short.addAction(
            open_short_user_home_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # extra profiles
        self._ui.profilesTableButtons.button(QtWidgets.QDialogButtonBox.Open).setText(
            "+"
        )
        self._ui.profilesTableButtons.button(
            QtWidgets.QDialogButtonBox.Open
        ).clicked.connect(self._profiles_add)
        self._ui.profilesTableButtons.button(QtWidgets.QDialogButtonBox.Close).setText(
            "-"
        )
        self._ui.profilesTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).clicked.connect(self._profiles_remove)
        self._ui.profilesTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).setEnabled(False)
        self._ui.profilesTable.itemSelectionChanged.connect(self._profiles_selection)
        # config
        self._ui.configPrintRunCommands.toggled.connect(
            self._config_toggle_printruncommands
        )
        self._ui.configRevisions.toggled.connect(self._config_toggle_revisions)
        # environments
        # - adding
        self._ui.envTableButtons.button(QtWidgets.QDialogButtonBox.Open).setText("+")
        self._ui.envTableButtons.button(
            QtWidgets.QDialogButtonBox.Open
        ).clicked.connect(self._env_add_plus)
        self._ui.envTableButtons.button(QtWidgets.QDialogButtonBox.Close).setText("-")
        self._ui.envTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).clicked.connect(self._env_add_minus)
        self._ui.envTableButtons.button(QtWidgets.QDialogButtonBox.Close).setEnabled(
            False
        )
        self._ui.envTable.itemSelectionChanged.connect(self._env_add_selection)
        self._ui.envTable.model().dataChanged.connect(self._env_add_changed)
        # - removing
        self._ui.envRemoveButtons.button(QtWidgets.QDialogButtonBox.Open).setText("+")
        self._ui.envRemoveButtons.button(
            QtWidgets.QDialogButtonBox.Open
        ).clicked.connect(self._env_remove_plus)
        self._ui.envRemoveButtons.button(QtWidgets.QDialogButtonBox.Close).setText("-")
        self._ui.envRemoveButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).clicked.connect(self._env_remove_minus)
        self._ui.envRemoveButtons.button(QtWidgets.QDialogButtonBox.Close).setEnabled(
            False
        )
        self._ui.envRemoveList.itemSelectionChanged.connect(self._env_remove_selection)
        self._ui.envRemoveList.model().dataChanged.connect(self._env_remove_changed)
        # remotes
        self._ui.remotesTableButtons.button(QtWidgets.QDialogButtonBox.Open).setText(
            "+"
        )
        self._ui.remotesTableButtons.button(
            QtWidgets.QDialogButtonBox.Open
        ).clicked.connect(self._remotes_add)
        self._ui.remotesTableButtons.button(QtWidgets.QDialogButtonBox.Close).setText(
            "-"
        )
        self._ui.remotesTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).clicked.connect(self._remotes_remove)
        self._ui.remotesTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).setEnabled(False)
        self._ui.remotesTable.itemSelectionChanged.connect(self._remotes_selection)
        self._ui.remotesTable.remotes_reordered.connect(self._remotes_reordered)
        self._ui.remotesTable.remote_enabled.connect(self._remote_enabled)
        # hooks
        self._ui.hooksTable.itemChanged.connect(self._hooks_item_changed)
        # operations
        self._ui.operations_installConfigButton.clicked.connect(
            self._operations_install_config
        )
        self._ui.operations_removeLocksButton.clicked.connect(
            self._operations_remove_locks
        )
        self._ui.operations_removeAllPackagesButton.clicked.connect(
            self._operations_remove_all_packages
        )
        self._ui.moveCacheButton.clicked.connect(self._operations_move_cache)
        self._ui.deleteCacheButton.clicked.connect(self._operations_delete_cache)
        self._ui.runConanCommandButton.clicked.connect(self._operations_run_command)
        # log
        self._ui.localCacheLog.hide()
        self._ui.localCacheLog.customContextMenuRequested.connect(
            self._log_context_menu
        )
        self._log_details = LogDetails(
            self._ui.localCacheLog, self._ui.localCacheLog, True, False, None
        )
        self._log_details.logging.connect(self._ui.localCacheLog.show)
        self._context = ConanContext(DEFAULT_CACHE_NAME, self._log_details)
        self._populate_cache_names(cache_name_to_open)
        self._modifications: typing.Dict[str, typing.Any] = {}
        self._modified.connect(self._on_modification)

    def _open_conan_user_home(self) -> None:
        reveal_on_filesystem(self._ui.conan_user_home.text())

    def _open_conan_user_home_short(self) -> None:
        reveal_on_filesystem(self._ui.conan_user_home_short.text())

    def _on_modification(self) -> None:
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Apply).setEnabled(
            bool(self._modifications)
        )

    def accept(self) -> None:
        if self._modifications:
            self._save_modifications()
        self._context.close()
        super().accept()

    def reject(self) -> None:
        if self._modifications:
            if (
                QtWidgets.QMessageBox.question(
                    self,
                    "Local cache modifications",
                    "Do you want to lose unsaved changes to the local cache?",
                )
                == QtWidgets.QMessageBox.StandardButton.No
            ):
                return
        self._context.close()
        super().reject()

    def _populate_cache_names(
        self, cache_name_to_select: typing.Optional[str] = None
    ) -> None:
        combo = self._ui.local_cache_names
        current_cache_name = combo.currentText()
        combo.currentTextChanged.connect(
            self._update_cache_details
        )  # TODO: should this be moved, as this fn is called multiple times
        with BlockSignals(combo):
            combo.clear()
            with AllNamedLocalCacheSettingsReader() as names:
                combo.addItems(names)
            combo.setCurrentIndex(-1)
        if cache_name_to_select:
            combo.setCurrentText(cache_name_to_select)
        else:
            # no preference, default or the previous selection
            combo.setCurrentText(current_cache_name or DEFAULT_CACHE_NAME)

    def _new_cache(self) -> None:
        dialog = NewLocalCacheWizard(self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            if dialog.switch_to_new_cache:
                self._populate_cache_names(dialog.cache_name)
            else:
                # slightly faster, as all cache details don't need to be refetched
                with BlockSignals(self._ui.local_cache_names):
                    self._populate_cache_names()
            self.cache_changed.emit(dialog.cache_name)

    def _update_cache_locations(
        self,
        cache_home_dir: typing.Optional[str],
        cache_short_home_dir: typing.Optional[str],
    ) -> None:
        is_windows = platform.system() == "Windows"
        if not is_windows:
            self._ui.loc_conan_user_home_short_label.hide()
            self._ui.conan_user_home_short.hide()
        if self._context.is_default:
            home_dir = QtCore.QStandardPaths.writableLocation(
                QtCore.QStandardPaths.HomeLocation
            )
            self._ui.conan_user_home.setText(home_dir)
            if is_windows:
                drive = QtCore.QStorageInfo(home_dir).name()
                self._ui.conan_user_home_short.setText(drive)
        else:
            assert cache_home_dir
            self._ui.conan_user_home.setText(cache_home_dir)
            if is_windows and cache_short_home_dir is not None:
                self._ui.conan_user_home_short.setText(cache_short_home_dir)
        with NamedLocalCacheSettingsReader(self._context.cache_name) as settings:
            recipe_uuids = settings.recipe_uuids
        self._ui.localCacheRecipeCount.setText(
            f"{len(recipe_uuids)} recipes associated with this cache"
        )

    def _update_cache_profiles(self, extra_profile_dirs: typing.Dict[str, str]) -> None:
        self._ui.profilesTable.setRowCount(0)
        for name, profile_dir in extra_profile_dirs.items():
            self._ui.profilesTable.add_key_value_pair(name, profile_dir)
        self._ui.profilesList.clear()
        for i, (profile_path, profile_text) in enumerate(
            self._context.get_list_of_profiles()
        ):
            self._ui.profilesList.addItem(str(profile_path))
            item = self._ui.profilesList.item(i)
            item.setData(QtCore.Qt.ToolTipRole, profile_text)  # type: ignore[arg-type]

    def _update_cache_config(self) -> None:
        with BlockSignals(self._ui.configPrintRunCommands) as blocked_widget:
            blocked_widget.setCheckState(
                QtCore.Qt.Checked
                if self._context.get_boolean_config(
                    ConanConfigBoolean.PRINT_RUN_COMMANDS, False
                )
                else QtCore.Qt.Unchecked
            )
        with BlockSignals(self._ui.configRevisions) as blocked_widget:
            blocked_widget.setCheckState(
                QtCore.Qt.Checked
                if self._context.get_boolean_config(ConanConfigBoolean.REVISIONS, False)
                else QtCore.Qt.Unchecked
            )

    def _update_cache_environment(
        self, env_added: typing.Dict[str, str], env_removed: typing.List[str]
    ) -> None:
        self._ui.envTable.setRowCount(0)
        for key, value in env_added.items():
            self._ui.envTable.add_key_value_pair(key, value)
        self._ui.envRemoveList.clear()
        for key in env_removed:
            item = QtWidgets.QListWidgetItem(key)
            self._ui.envRemoveList.addItem(item)

    def _update_cache_remotes(self) -> None:
        self._ui.remotesTable.setRowCount(0)
        for remote in self._context.get_remotes_list():
            self._ui.remotesTable.add_remote(remote)

    def _update_cache_hooks(self) -> None:
        self._ui.hooksTable.setRowCount(0)
        hooks = self._context.get_hooks_list()
        self._ui.hooksTable.setRowCount(len(hooks))
        for i, hook in enumerate(hooks):
            enabled_item = QtWidgets.QTableWidgetItem()
            enabled_item.setFlags(
                QtCore.Qt.ItemIsEnabled  # type: ignore
                | QtCore.Qt.ItemIsSelectable
                | QtCore.Qt.ItemIsUserCheckable
            )
            enabled_item.setCheckState(
                QtCore.Qt.Checked if hook.enabled else QtCore.Qt.Unchecked
            )
            name_item = QtWidgets.QTableWidgetItem(hook.path)
            name_item.setFlags(
                QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable  # type: ignore
            )
            with BlockSignals(self._ui.hooksTable) as blocked_widget:
                blocked_widget.setItem(
                    i,
                    ManageLocalCachesDialog._HooksTableColumnIndex.ENABLED,
                    enabled_item,
                )
                blocked_widget.setItem(
                    i, ManageLocalCachesDialog._HooksTableColumnIndex.PATH, name_item
                )

    def _update_cache_details(self, cache_name: str) -> None:
        self._context.change_cache(cache_name)
        with NamedLocalCacheSettingsReader(cache_name) as settings:
            home_dir = settings.home_dir.resolve()
            short_home_dir = settings.short_home_dir.resolve()
            env_added = settings.environment_added.resolve()
            env_removed = settings.environment_removed.resolve()
            extra_profile_dirs = settings.extra_profile_directories.resolve()
            recipe_uuids = settings.recipe_uuids
        self._update_cache_locations(home_dir, short_home_dir)
        self._update_cache_profiles(extra_profile_dirs)
        self._update_cache_config()
        self._update_cache_environment(env_added, env_removed)
        self._update_cache_remotes()
        self._update_cache_hooks()
        can_move = not self._context.is_default
        self._ui.moveCacheButton.setEnabled(can_move)
        can_delete = not self._context.is_default and not recipe_uuids
        self._ui.deleteCacheButton.setEnabled(can_delete)

    def _save_modifications(self) -> None:
        assert self._modifications
        if "Profiles" in self._modifications or "Env" in self._modifications:
            settings_lc = NamedLocalCacheSettings()
            if "Profiles" in self._modifications:
                if "Added" in self._modifications["Profiles"]:
                    for extra_dir in self._modifications["Profiles"]["Added"]:
                        settings_lc.append_extra_profile_directories(extra_dir)
                if "Removed" in self._modifications["Profiles"]:
                    for name in self._modifications["Profiles"]["Removed"]:
                        settings_lc.remove_extra_profile_directories(name)
            if "Env" in self._modifications and self._modifications["Env"].has_change:
                with NamedLocalCacheSettingsReader(
                    self._context.cache_name
                ) as settings:
                    existing_removed = settings.environment_removed.resolve()
                settings_lc.sync_env_changes(
                    existing_removed, self._modifications["Env"]
                )
            NamedLocalCacheSettingsWriter(self._context.cache_name).sync(settings_lc)
            if "Profiles" in self._modifications:
                del self._modifications["Profiles"]
            if "Env" in self._modifications:
                del self._modifications["Env"]
        if "Config" in self._modifications:
            for config, value in self._modifications["Config"].items():
                self._context.set_boolean_config(config, value)
            del self._modifications["Config"]
        if "Remotes" in self._modifications:
            # brute force write everything back to settings
            self._context.remotes_sync(self._ui.remotesTable.remotes_list())
            if "Add" in self._modifications["Remotes"]:
                with RecentConanRemotesSettingsReader() as settings_remotes:
                    urls = settings_remotes.urls.resolve()
                for remote in self._modifications["Remotes"]["Add"]:
                    if remote.url not in urls:
                        urls.append(remote.url)
                settings_remotes.urls = urls  # type: ignore[assignment]
                RecentConanRemotesSettingsWriter().sync(settings_remotes)
            del self._modifications["Remotes"]
        if "Hooks" in self._modifications:
            self._context.hooks_sync(self._modifications["Hooks"])
            del self._modifications["Hooks"]
        assert not self._modifications
        self._modified.emit()
        # need to refresh if it was Apply
        if self.sender() == self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Apply):
            self._update_cache_details(self._context.cache_name)
        self.cache_changed.emit(self._context.cache_name)

    def _remotes_add(self) -> None:
        dialog = AddRemoteDialog(self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            self._ui.remotesTable.add_remote(dialog.new_remote)
            if "Remotes" not in self._modifications:
                self._modifications["Remotes"] = {}
            if "Add" not in self._modifications:
                self._modifications["Remotes"]["Add"] = []
            self._modifications["Remotes"]["Add"].append(dialog.new_remote)
            self._modified.emit()

    def _remotes_remove(self) -> None:
        remote_removed = self._ui.remotesTable.remove_selected()
        if "Remotes" not in self._modifications:
            self._modifications["Remotes"] = {}
        if "Add" in self._modifications["Remotes"]:
            index = next(
                (
                    i
                    for i, remote in enumerate(self._modifications["Remotes"]["Add"])
                    if remote.name == remote_removed.name
                ),
                -1,
            )
            if index >= 0:
                del self._modifications["Remotes"]["Add"][index]
                if not self._modifications["Remotes"]["Add"]:
                    del self._modifications["Remotes"]["Add"]
                if not self._modifications["Remotes"]:
                    del self._modifications["Remotes"]
                if (
                    "Remotes" in self._modifications
                    and "Reordered" in self._modifications["Remotes"]
                ):
                    # has the removal of the addition now put the
                    # table back into it's original order?
                    if self._ui.remotesTable.same(self._context.get_remotes_list()):
                        del self._modifications["Remotes"]["Reordered"]
                        if not self._modifications["Remotes"]:
                            del self._modifications["Remotes"]
        else:
            if "Remove" not in self._modifications["Remotes"]:
                self._modifications["Remotes"]["Remove"] = []
            self._modifications["Remotes"]["Remove"].append(remote_removed.name)
        self._modified.emit()

    def _remotes_selection(self) -> None:
        self._ui.remotesTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).setEnabled(bool(self._ui.remotesTable.selectedRanges()))

    def _remotes_reordered(self) -> None:
        if self._ui.remotesTable.same(self._context.get_remotes_list()):
            del self._modifications["Remotes"]["Reordered"]
            if not self._modifications["Remotes"]:
                del self._modifications["Remotes"]
        else:
            if "Remotes" not in self._modifications:
                self._modifications["Remotes"] = {}
            if "Reordered" not in self._modifications["Remotes"]:
                self._modifications["Remotes"]["Reordered"] = True
        self._modified.emit()

    def _remote_enabled(self, name: str, state: QtCore.Qt.CheckState) -> None:
        remotes = self._context.get_remotes_list()
        try:
            remote = next(item for item in remotes if item.name == name)
            if remote.enabled == bool(state):
                self._modifications["Remotes"]["Toggled"].remove(name)
                if not self._modifications["Remotes"]["Toggled"]:
                    del self._modifications["Remotes"]["Toggled"]
                if not self._modifications["Remotes"]:
                    del self._modifications["Remotes"]
            else:
                if "Remotes" not in self._modifications:
                    self._modifications["Remotes"] = {}
                if "Toggled" not in self._modifications["Remotes"]:
                    self._modifications["Remotes"]["Toggled"] = []
                self._modifications["Remotes"]["Toggled"].append(name)
        except StopIteration:
            assert "Remotes" in self._modifications
            assert "Add" in self._modifications["Remotes"]
            remote = next(
                item
                for item in self._modifications["Remotes"]["Add"]
                if item.name == name
            )
            # this was simply to check that the modifications contained the remote
            # toggled since the table holds the truth, not the modifications
        self._modified.emit()

    def _hooks_item_changed(self, item: QtWidgets.QTableWidgetItem) -> None:
        if item.column() == ManageLocalCachesDialog._HooksTableColumnIndex.ENABLED:
            path_item = self._ui.hooksTable.item(
                item.row(), ManageLocalCachesDialog._HooksTableColumnIndex.PATH
            )
            hook_path = path_item.text().strip()
            item_enabled = bool(item.checkState())
            hooks = self._context.get_hooks_list()
            hook = next(item for item in hooks if item.has_path(hook_path))
            assert hook
            if item_enabled == hook.enabled:
                hook_unchanged = next(
                    item
                    for item in self._modifications["Hooks"]
                    if item.has_path(hook_path)
                )
                self._modifications["Hooks"].remove(hook_unchanged)
                if not self._modifications["Hooks"]:
                    del self._modifications["Hooks"]
            else:
                if "Hooks" not in self._modifications:
                    self._modifications["Hooks"] = []
                self._modifications["Hooks"].append(ConanHook(hook_path, item_enabled))
            self._modified.emit()

    def _operations_install_config(self) -> None:
        if (
            InstallConfigDialog(self._context, self).exec_()
            == QtWidgets.QDialog.Accepted
        ):
            cache_name = self._context.cache_name
            self._update_cache_details(cache_name)
            self.cache_changed.emit(cache_name)

    def _operations_remove_locks(self) -> None:
        RemoveLocksDialog(self._context, self).exec_()

    def _operations_remove_all_packages(self) -> None:
        result = QtWidgets.QMessageBox.question(
            self,
            "Package deletion",
            "Are you sure you want to delete all packages from this local cache?",
        )
        if result == QtWidgets.QMessageBox.StandardButton.Yes:
            RemoveAllPackagesDialog(self._context, self).exec_()

    def _operations_move_cache(self) -> None:
        if (
            MoveLocalCacheDialog(self._context, self).exec_()
            == QtWidgets.QDialog.Accepted
        ):
            self._update_cache_details(self._context.cache_name)

    @staticmethod
    def _delete_tree_with_readonly_files(dir_path: str) -> None:
        if not os.path.isdir(dir_path):
            return

        def make_writeable(action: typing.Any, name: str, exc: typing.Any) -> None:
            # pylint: disable=unused-argument
            if exc[0] is PermissionError:
                os.chmod(name, stat.S_IWRITE)
                os.remove(name)
            else:
                raise exc[1]

        shutil.rmtree(dir_path, onerror=make_writeable)

    def _operations_delete_cache(self) -> None:
        cache_name = self._context.cache_name
        result = QtWidgets.QMessageBox.question(
            self,
            "Local cache deletion",
            f"Are you sure you want to delete the local cache {cache_name}?",
        )
        if result == QtWidgets.QMessageBox.StandardButton.No:
            return
        with NamedLocalCacheSettingsReader(cache_name) as settings:
            home_dir = settings.home_dir.resolve()
            short_home_dir = settings.short_home_dir.resolve()

        assert home_dir  # because it's non-default
        conan_home_dir = os.path.join(home_dir, ".conan")
        # since this is destructive, check again
        result = QtWidgets.QMessageBox.question(
            self,
            "Local cache deletion",
            f"Please confirm you want to delete the directories {conan_home_dir} and "
            f"{short_home_dir}?"
            if short_home_dir
            else f"Please confirm you want to delete the directory {conan_home_dir}?",
        )
        if result == QtWidgets.QMessageBox.StandardButton.No:
            return
        ManageLocalCachesDialog._delete_tree_with_readonly_files(conan_home_dir)
        if short_home_dir:
            ManageLocalCachesDialog._delete_tree_with_readonly_files(short_home_dir)
        NamedLocalCacheDeleter().delete(cache_name)
        self._ui.local_cache_names.removeItem(
            self._ui.local_cache_names.findText(cache_name)
        )
        self._ui.local_cache_names.setCurrentIndex(0)  # default
        self.cache_changed.emit(cache_name)

    def _operations_run_command(self) -> None:
        RunConanCommandDialog(self._context.cache_name, self).exec_()

    def _profiles_add(self) -> None:
        dialog = AddExtraProfileDirectoryDialog(self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            extra = dialog.extra
            self._ui.profilesTable.add_key_value_pair(extra.name, extra.directory)
            if "Profiles" not in self._modifications:
                self._modifications["Profiles"] = {}
            if "Added" not in self._modifications["Profiles"]:
                self._modifications["Profiles"]["Added"] = []
            self._modifications["Profiles"]["Added"].append(extra)
            self._modified.emit()

    def _profiles_remove(self) -> None:
        name_removed = self._ui.profilesTable.remove_selected()
        if "Profiles" not in self._modifications:
            self._modifications["Profiles"] = {}
        if "Added" in self._modifications["Profiles"]:
            try:
                profile_add_removed = next(
                    item
                    for item in self._modifications["Profiles"]["Added"]
                    if item.name == name_removed
                )
                self._modifications["Profiles"]["Added"].remove(profile_add_removed)
                if not self._modifications["Profiles"]["Added"]:
                    del self._modifications["Profiles"]["Added"]
                if not self._modifications["Profiles"]:
                    del self._modifications["Profiles"]
                self._modified.emit()
                return
            except StopIteration:
                pass
        if "Removed" not in self._modifications["Profiles"]:
            self._modifications["Profiles"]["Removed"] = []
        self._modifications["Profiles"]["Removed"].append(name_removed)
        self._modified.emit()

    def _profiles_selection(self) -> None:
        self._ui.profilesTableButtons.button(
            QtWidgets.QDialogButtonBox.Close
        ).setEnabled(bool(self._ui.profilesTable.selectedRanges()))

    def _config_toggle(self, checked: bool, config: ConanConfigBoolean) -> None:
        if checked == self._context.get_boolean_config(config, False):
            del self._modifications["Config"][config]
            if not self._modifications["Config"]:
                del self._modifications["Config"]
        else:
            if "Config" not in self._modifications:
                self._modifications["Config"] = {}
            self._modifications["Config"][config] = checked
        self._modified.emit()

    def _config_toggle_printruncommands(self, checked: bool) -> None:
        self._config_toggle(checked, ConanConfigBoolean.PRINT_RUN_COMMANDS)

    def _config_toggle_revisions(self, checked: bool) -> None:
        self._config_toggle(checked, ConanConfigBoolean.REVISIONS)

    def _env_add_plus(self) -> None:
        dialog = AddEnvironmentDialog(self._context, self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            new_env = dialog.environment_variable
            self._ui.envTable.add_key_value_pair(new_env.key, new_env.value)
            if "Env" not in self._modifications:
                self._modifications["Env"] = _EnvChangeManagement()
            self._modifications["Env"].key_required(new_env.key, new_env.value)
            self._modified.emit()

    def _env_add_minus(self) -> None:
        key_removed = self._ui.envTable.remove_selected()
        if "Env" not in self._modifications:
            self._modifications["Env"] = _EnvChangeManagement()
        self._modifications["Env"].key_not_required(key_removed)
        if not self._modifications["Env"].has_change:
            del self._modifications["Env"]
        self._modified.emit()

    def _env_add_selection(self) -> None:
        self._ui.envTableButtons.button(QtWidgets.QDialogButtonBox.Close).setEnabled(
            bool(self._ui.envTable.selectedRanges())
        )

    def _env_add_changed(
        self,
        top_left: QtCore.QModelIndex,
        bottom_right: QtCore.QModelIndex,
        roles: typing.List[int],
    ) -> None:
        assert top_left == bottom_right
        row_index = top_left.row()
        new_value = top_left.data(0)
        with NamedLocalCacheSettingsReader(self._context.cache_name) as settings:
            key, prev_value = settings.environment_added_at(row_index)
        if new_value == prev_value:
            return
        if "Env" not in self._modifications:
            self._modifications["Env"] = _EnvChangeManagement()
        self._modifications["Env"].key_required(key, new_value)
        self._modified.emit()

    def _env_remove_plus(self) -> None:
        dialog = RemoveEnvironmentDialog(self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            new_item = QtWidgets.QListWidgetItem(dialog.name)
            self._ui.envRemoveList.addItem(new_item)
            if "Env" not in self._modifications:
                self._modifications["Env"] = _EnvChangeManagement()
            self._modifications["Env"].key_never_required(dialog.name)
            self._modified.emit()

    def _env_remove_minus(self) -> None:
        items_selected = self._ui.envRemoveList.selectedItems()
        assert len(items_selected) == 1
        item_removed = self._ui.envRemoveList.takeItem(
            self._ui.envRemoveList.row(items_selected[0])
        )
        key_removed = item_removed.text()
        if "Env" not in self._modifications:
            self._modifications["Env"] = _EnvChangeManagement()
        self._modifications["Env"].key_not_never_required(key_removed)
        if not self._modifications["Env"].has_change:
            del self._modifications["Env"]
        self._modified.emit()

    def _env_remove_selection(self) -> None:
        self._ui.envRemoveButtons.button(QtWidgets.QDialogButtonBox.Close).setEnabled(
            bool(self._ui.envRemoveList.selectedItems())
        )

    def _env_remove_changed(
        self,
        top_left: QtCore.QModelIndex,
        bottom_right: QtCore.QModelIndex,
        roles: typing.List[int],
    ) -> None:
        assert top_left == bottom_right
        row_index = top_left.row()
        new_value = top_left.data(0)
        # TODO: environment_added_at not correct, as not returning a pair
        with NamedLocalCacheSettingsReader(self._context.cache_name) as settings:
            key, prev_value = settings.environment_added_at(row_index)
        if new_value == prev_value:
            return
        assert "Env" in self._modifications
        assert isinstance(self._modifications["Env"], _EnvChangeManagement)
        self._modifications["Env"].key_never_required(key)
        self._modified.emit()

    def _log_context_menu(self, position: QtCore.QPoint) -> None:
        menu = self.sender().createStandardContextMenu(position)
        menu.addSeparator()
        clear_action = QAction("Clear", self)
        clear_action.triggered.connect(self._clear_log)
        menu.addAction(clear_action)
        menu.exec_(self.sender().viewport().mapToGlobal(position))

    def _clear_log(self) -> None:
        self._ui.localCacheLog.clear()
        self._ui.localCacheLog.hide()



#!/usr/bin/env python3

"""
A context object wrapping the Conan API
"""

from contextlib import contextmanager
import logging
import os
import pathlib
import typing

from qtpy import QtCore, QtWidgets

from cruiz.exceptions import RecipeInspectionError

import cruiz.interop.dependencygraph
from cruiz.interop.packagenode import PackageNode
from cruiz.interop.pod import ConanRemote, ConanHook
from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.searchrecipesparameters import SearchRecipesParameters
from cruiz.interop.reciperevisionsparameters import RecipeRevisionsParameters
from cruiz.interop.packageidparameters import PackageIdParameters
from cruiz.interop.packagerevisionsparameters import PackageRevisionsParameters
from cruiz.interop.packagebinaryparameters import PackageBinaryParameters
from cruiz.recipe.logs.command import CommandListWidgetItem
from cruiz.settings.managers.namedlocalcache import NamedLocalCacheSettingsReader

import cruiz.workers.create
import cruiz.workers.imports
import cruiz.workers.install
import cruiz.workers.source
import cruiz.workers.build
import cruiz.workers.package
import cruiz.workers.exportpackage
import cruiz.workers.testpackage
import cruiz.workers.cmakebuildtool
import cruiz.workers.deletecmakecache
import cruiz.workers.removepackage
import cruiz.workers.removeallpackages
import cruiz.workers.removelocks
import cruiz.workers.configinstall
import cruiz.workers.remotesearch
import cruiz.workers.reciperevisions
import cruiz.workers.packagedetails
import cruiz.workers.packagerevisions
import cruiz.workers.packagebinary
import cruiz.workers.arbitrary

from cruiz.recipe.logs.command import RecipeCommandHistoryWidget
from cruiz.constants import DEFAULT_CACHE_NAME

from .conaninvocation import ConanInvocation
from .metarequestconaninvocation import MetaRequestConanInvocation
from .conanenv import get_conan_env
from .conanconf import ConanConfigBoolean
from .logdetails import LogDetails


logger = logging.getLogger(__name__)


# copied from distutils.url.strtobool and modified
def _strtobool(val: str) -> bool:
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError("invalid truth value %r" % (val,))


# pylint: disable=too-many-public-methods
class ConanContext(QtCore.QObject):
    """
    Context for invoking Conan instances.
    Some are for long running commands, with a background thread to process information.
    Others are for just quering meta data.
    """

    completed = QtCore.Signal(object, Exception)
    cancelled = QtCore.Signal()

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(
        self,
        cache_name: str,
        log_details: LogDetails,
    ) -> None:
        logger.debug("+=%d", id(self))
        super().__init__(None)
        self.command_history_widget: typing.Optional[RecipeCommandHistoryWidget] = None
        self._log_details = log_details
        self._invocations: typing.List[ConanInvocation] = []
        self._configure_to_local_cache(cache_name)

    def close(self) -> None:
        """
        Close the context and any resources associated with it
        """
        try:
            assert not self.is_busy
            self._meta_invocation.close()
            self._meta_invocation.deleteLater()
            del self._meta_invocation
        except AttributeError:
            # don't double close this
            pass

    def change_cache(self, cache_name: str) -> None:
        """
        Change the local cache used by this context.
        """
        # don't even try to check for a name no-op if the cache has already been closed
        if hasattr(self, "_meta_invocation"):
            if cache_name == self.cache_name:
                return
            assert not self.is_busy
            self.close()
        self._configure_to_local_cache(cache_name)

    def _configure_to_local_cache(self, cache_name: str) -> None:
        self.cache_name = cache_name
        self._meta_invocation = MetaRequestConanInvocation(
            self,
            cache_name,
            self._log_details,
        )

    def _start_invocation(
        self,
        parameters: typing.Union[
            CommandParameters,
            SearchRecipesParameters,
            RecipeRevisionsParameters,
            PackageIdParameters,
            PackageRevisionsParameters,
            PackageBinaryParameters,
        ],
        log_details: LogDetails,
        command_toolbar: typing.Optional[QtWidgets.QWidget],
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
        enable_history: bool = True,
    ) -> None:
        added_environment, removed_environment = get_conan_env(self.cache_name)
        parameters.added_environment.update(added_environment)
        parameters.removed_environment.extend(removed_environment)
        instance = ConanInvocation(self._log_details)
        instance.completed.connect(self._completed_invocation)
        instance.finished.connect(self._finished_invocation)
        if command_toolbar:
            instance.finished.connect(command_toolbar.command_ended)
        instance.invoke(parameters, log_details, continuation)
        self._invocations.append(instance)
        if self.command_history_widget is not None and enable_history:
            # TODO: prefer adding to a model?
            assert isinstance(parameters, CommandParameters)
            item = CommandListWidgetItem(parameters)
            self.command_history_widget.addItem(item)

    def _completed_invocation(self, success: typing.Any, exception: typing.Any) -> None:
        # pylint: disable=unused-argument
        logger.debug("COMPLETED invocation (%d)", id(self.sender()))
        self.sender().close()

    def _finished_invocation(self) -> None:
        # can only free the reference to the ConanInvocation once the message processing
        # thread has been destroyed
        instance = self.sender()
        logger.debug("FINISHED invocation (%d)", id(instance))
        instance.deleteLater()
        assert isinstance(instance, ConanInvocation)
        self._invocations.remove(instance)

    def conancommand(
        self,
        params: CommandParameters,
        log_details: LogDetails,
        command_toolbar: typing.Optional[QtWidgets.QWidget],
        continuation: typing.Optional[
            typing.Callable[[typing.Any, typing.Any], None]
        ] = None,
    ) -> None:
        """
        Run a conan command, with parameters as provided, and a continuation
        if further processing is needed
        """
        self._start_invocation(params, log_details, command_toolbar, continuation)

    def cmakebuildcommand(
        self,
        params: CommandParameters,
        log_details: LogDetails,
        command_toolbar: typing.Optional[QtWidgets.QWidget],
        continuation: typing.Optional[
            typing.Callable[[typing.Any, typing.Any], None]
        ] = None,
    ) -> None:
        """
        Invoke the CMake build tool on the build folder of the recipe
        """
        self._start_invocation(
            params, log_details, command_toolbar, continuation, enable_history=False
        )

    # pylint: disable=pointless-string-statement
    """
    def _make_symbolic_link(self, source: pathlib.Path, link: pathlib.Path) -> None:
        if link.exists():
            assert link.is_symlink()
            return
        if not link.parent.exists():
            os.makedirs(link.parent)
        # on Windows, os.symlink requires elevated privileges, or in developer mode
        # from Py3.8 although I cannot get the developer mode to work
        # additionally though, on Windows, I can't then open the symbolic link file
        # from Explorer even though a cmd dir command does show what its linked to
        os.symlink(source, link)
    """

    @staticmethod
    def _make_hard_link(source: pathlib.Path, link: pathlib.Path) -> None:
        if link.exists():
            if os.path.samefile(source, link):
                return
            # the inode may have changed, so just remove the old link, and recreate
            os.unlink(link)
        if not link.parent.exists():
            os.makedirs(link.parent)
        os.link(source, link)

    def editable_add(
        self,
        ref: str,
        recipe_path: pathlib.Path,
        config_subdir: str,
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
    ) -> None:
        """
        Run 'conan editable add'
        """
        # this path has to match that in BaseCommand._get_paths
        package_dir = recipe_path.parent / "_edit_" / config_subdir / "package"
        target_symlink = package_dir / recipe_path.name
        # since the path changes with profile, always try to create a link
        # if it doesn't already exist
        ConanContext._make_hard_link(recipe_path, target_symlink)

        result, exception = self._meta_invocation.request_data(
            "editable_add", {"ref": ref, "path": str(target_symlink)}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Editable add failed") from exception
        if continuation:
            continuation(result, exception)

    def editable_remove(
        self,
        ref: str,
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
    ) -> None:
        """
        Run 'conan editable remove'
        """
        result, exception = self._meta_invocation.request_data(
            "editable_remove", {"ref": ref}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Editable remove failed") from exception
        if continuation:
            continuation(result, exception)

    def remove_all_packages(
        self,
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
    ) -> None:
        """
        Run 'conan remove [-f] *'
        """
        params = CommandParameters(
            "removeallpackages", cruiz.workers.removeallpackages.invoke
        )
        params.force = True
        self._start_invocation(params, self._log_details, None, continuation)

    def remove_local_cache_locks(
        self,
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
    ) -> None:
        """
        Run 'conan remove --locks' to remove all lock files from the local cache
        """
        params = CommandParameters("removelocks", cruiz.workers.removelocks.invoke)
        self._start_invocation(params, self._log_details, None, continuation)

    def install_config(
        self,
        params: CommandParameters,
        continuation: typing.Optional[
            typing.Callable[[typing.Any, typing.Any], None]
        ] = None,
    ) -> None:
        """
        Run 'conan config install <URI> [--args -b branch] [-s source] [-t target]'
        """
        self._start_invocation(params, self._log_details, None, continuation)

    def remotes_sync(self, remotes: typing.List[ConanRemote]) -> None:
        """
        Sync the remotes to the given list.
        """
        _, exception = self._meta_invocation.request_data(
            "remotes_sync", {"remotes": remotes}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Sync remotes failed") from exception

    def get_remotes_list(self) -> typing.List[ConanRemote]:
        """
        Equivalent to 'conan remote list'
        """
        remotes_list, exception = self._meta_invocation.request_data("remotes_list")
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get remote list failed") from exception
        else:
            assert isinstance(remotes_list, list)
        return remotes_list

    def get_profile_meta(
        self, profile_name: str
    ) -> typing.Dict[str, typing.Dict[str, str]]:
        """
        Get the meta information for a named profile
        """
        result, exception = self._meta_invocation.request_data(
            "profile_meta", {"name": profile_name}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get profile meta failed") from exception
        else:
            assert isinstance(result, dict)
        return result

    def get_package_details(
        self,
        params: typing.Union[
            SearchRecipesParameters,
            RecipeRevisionsParameters,
            PackageIdParameters,
            PackageRevisionsParameters,
            PackageBinaryParameters,
        ],
        log_details: LogDetails,
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
    ) -> None:
        """
        Perform one of the actions to get package details from a remote.
        """
        self._start_invocation(params, log_details, None, continuation)

    def get_package_directory(self, pkgdata: PackageNode) -> str:
        """
        Get the package directory for the specified package.
        """
        package_dir, exception = self._meta_invocation.request_data(
            "package_dir",
            {
                "ref": pkgdata.reference,
                "package_id": pkgdata.package_id,
                "revision": pkgdata.recipe_revision,
                "short_paths": pkgdata.short_paths,
            },
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get package directory failed") from exception
        else:
            assert isinstance(package_dir, str)
        return package_dir

    def cancel(self) -> None:
        """
        Cancel any current running worker thread.
        """
        if self._invocations:
            for invocation in self._invocations:
                invocation.cancel()
                invocation.close()
            self.cancelled.emit()

    def conan_version(self) -> str:
        # pylint: disable=no-self-use
        """
        Get the Conan version
        """
        version, exception = self._meta_invocation.request_data("version")
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get Conan version failed") from exception
        else:
            assert isinstance(version, str)
        return version

    def profiles_dir(self) -> pathlib.Path:
        """
        Get the directory containing the profiles in the local cache.
        """
        profiles_dir, exception = self._meta_invocation.request_data("profiles_dir")
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get profiles directory failed") from exception
        else:
            assert isinstance(profiles_dir, str)
        return pathlib.Path(profiles_dir)

    def all_profile_dirs(self) -> typing.List[pathlib.Path]:
        """
        Get a list of all profile directories inspected.
        """
        profile_dirs = [self.profiles_dir()]
        with NamedLocalCacheSettingsReader(self.cache_name) as settings:
            extra_profile_dirs = settings.extra_profile_directories.resolve()
        for _, dir_path in extra_profile_dirs.items():
            profile_dirs.append(pathlib.Path(dir_path))
        return profile_dirs

    def default_profile_path(self) -> str:
        """
        Get the default profile path from the Conan object.
        """
        default_path, exception = self._meta_invocation.request_data(
            "default_profile_path"
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get default profile path failed") from exception
        else:
            assert isinstance(default_path, str)
        return default_path

    def default_profile_filename(self) -> str:
        """
        Query Conan for the default profile filename.
        """
        return str(
            pathlib.Path(self.default_profile_path()).relative_to(self.profiles_dir())
        )

    def get_list_of_profiles(self) -> typing.List[typing.Tuple[pathlib.Path, str]]:
        """
        Return a list of all profiles in this context.
        """
        profile_paths: typing.List[typing.Tuple[pathlib.Path, str]] = []
        profiles_dir = QtCore.QDir(str(self.profiles_dir()))
        for profile in profiles_dir.entryList(  # type: ignore
            filters=QtCore.QDir.Files
        ):
            # local cache profiles are listed relative
            path = pathlib.Path(profiles_dir.filePath(profile))
            text = path.read_text()
            profile_paths.append((pathlib.Path(path.name), text))
        assert profile_paths
        with NamedLocalCacheSettingsReader(self.cache_name) as settings:
            extra_profile_dirs = settings.extra_profile_directories.resolve()
        for _, extra_profile_dir in extra_profile_dirs.items():
            profiles_dir = QtCore.QDir(extra_profile_dir)
            for profile in profiles_dir.entryList(  # type: ignore
                filters=QtCore.QDir.Files
            ):
                # extra profiles are listed absolute
                path = pathlib.Path(profiles_dir.filePath(profile))
                text = path.read_text()
                profile_paths.append((path, text))
        return profile_paths

    def run_any_command(
        self,
        params: CommandParameters,
        continuation: typing.Callable[[typing.Any, typing.Any], None],
    ) -> None:
        """
        Run an arbitrary Conan command in the local cache.
        """

        self._start_invocation(params, self._log_details, None, continuation)

    # TODO: remove editables
    def get_editables_list(self) -> typing.Dict[str, str]:
        """
        Equivalent to 'conan editable list'
        """
        editable_dict, exception = self._meta_invocation.request_data("editable_list")
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get editables list failed") from exception
        else:
            assert isinstance(editable_dict, dict)
        return editable_dict

    def inspect_recipe(
        self, recipe_path: pathlib.Path, propagate_errors: bool = False
    ) -> typing.Dict[str, str]:
        """
        Equivalent to 'conan inspect'
        """
        results, exception = self._meta_invocation.request_data(
            "inspect_recipe", {"path": recipe_path}
        )
        if exception:
            lines = str(exception).splitlines()
            html = ""
            for line in lines:
                stripped_line = line.lstrip()
                num_leading_spaces = len(line) - len(stripped_line)
                html += "&nbsp;" * num_leading_spaces + stripped_line + "<br>"
            self._log_details.stderr(
                f"Exception raised from running command:<br>"
                f"<font color='red'>{html}</font><br>"
            )
            if propagate_errors:
                raise RecipeInspectionError()
            return {}
        assert isinstance(results, dict)
        return results

    def get_hooks_list(self) -> typing.List[ConanHook]:
        """
        No equivalent on the command line, but lists all hooks in the config and their
        enabled states
        """
        hooks_list, exception = self._meta_invocation.request_data("get_hooks")
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Get hooks list failed") from exception
        else:
            assert isinstance(hooks_list, list)
        return hooks_list

    def hooks_sync(self, hook_changes: typing.List[ConanHook]) -> None:
        """
        Equivalent to either conan config set <hook> or conan config rm <hook>
        """
        _, exception = self._meta_invocation.request_data(
            "hooks_sync", {"hooks": hook_changes}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Syncing hooks failed") from exception

    def enable_hook(self, hook: str, enabled: bool) -> None:
        """
        Equivalent to either conan config set <hook> or conan config rm <hook>
        """
        _, exception = self._meta_invocation.request_data(
            "enable_hook", {"hook": hook, "hook_enabled": enabled}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Enable/disable hook failed") from exception

    def get_cmake_generator(self) -> str:
        """
        Equivalent to conan config get general.cmake_generator
        """
        generator, exception = self._meta_invocation.request_data(
            "get_cmake_generator", {}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception("Failed to get CMake generator") from exception
        return generator

    def get_conandata(
        self, recipe_path: pathlib.Path
    ) -> typing.Dict[str, typing.Dict[str, str]]:
        """
        Fetch the YAML dictionary of the conandata.yml beside the recipe.
        """
        conandata, exception = self._meta_invocation.request_data(
            "get_conandata", {"path": recipe_path}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception(
                    "Failed to get conandata YAML dict for recipe"
                ) from exception
        return conandata

    def get_boolean_config(
        self, config: ConanConfigBoolean, default_value: bool
    ) -> bool:
        """
        Equivalent to conan config get, with the specified config key.
        If the specified config key is not in the configuration,
        return the default value.
        """
        config_value, exception = self._meta_invocation.request_data(
            "get_config", {"config": config.value}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception(
                    f"Failed to get local cache config for '{config.value}'"
                ) from exception
        if config_value is not None:
            assert isinstance(config_value, str)
            config_value = _strtobool(config_value)
        return config_value or default_value

    def set_boolean_config(self, config: ConanConfigBoolean, value: bool) -> None:
        """
        Equivalent to conan config set, with the specified config key and value.
        """
        _, exception = self._meta_invocation.request_data(
            "set_config", {"config": config.value, "value": str(value)}
        )
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception(
                    f"Failed to set local cache config for '{config.value}'"
                ) from exception

    @property
    def is_default(self) -> bool:
        """
        Does this context refer to the 'default' local cache?
        """
        return self.cache_name == DEFAULT_CACHE_NAME

    def get_conan_config_environment_variables(self) -> typing.Dict[str, str]:
        """
        Get a dictionary of all Conan environment variables.
        This doesn't get all of them, but at least some.
        """
        envvars, exception = self._meta_invocation.request_data("get_config_envvars")
        if exception:
            if self.parent():
                self.parent().record_exception(exception)
            else:
                raise Exception(
                    "Failed to get Conan config environment variables"
                ) from exception
        return envvars

    @property
    def is_busy(self) -> bool:
        """
        Is the context busy running any Conan commands?
        Explicitly or via its meta commands
        """
        return bool(self._invocations) or self._meta_invocation.active


@contextmanager
def managed_conan_context(
    cache_name: str, log_details: LogDetails
) -> typing.Generator[ConanContext, None, None]:
    """
    Context manager for uses where all API calls on the ConanContext are guaranteed
    to be synchronous.
    """
    context = ConanContext(cache_name, log_details)
    try:
        yield context
    finally:
        context.close()



from contextlib import contextmanager
import os
import subprocess
import typing

from setuptools import setup, find_packages
import setuptools.command.build_py
import setuptools.command.sdist

from cruiz.version import get_version


def _run_command(args: typing.List[str]) -> None:
    print(f"Running {' '.join(args)}", flush=True)
    subprocess.check_call(args)


@contextmanager
def _temp_resource_generation(
    callback: typing.Callable[[typing.List[str]], None]
) -> typing.Generator[None, None, None]:
    paths_written: typing.List[str] = []
    callback(paths_written)
    try:
        yield
    finally:
        for path in paths_written:
            os.unlink(path)


def _compile_qrc(paths_written: typing.List[str]) -> None:
    def _compile_pyside_resources(ver: int) -> str:
        if not os.path.isdir(f"cruiz/pyside{ver}"):
            os.makedirs(f"cruiz/pyside{ver}")
        dst_file = f"cruiz/pyside{ver}/resources.py"
        _run_command(
            [
                f"pyside{ver}-rcc",
                "--verbose",
                "-o",
                dst_file,
                "cruiz/resources/cruizres.qrc",
            ]
        )
        return dst_file

    paths_written.append(_compile_pyside_resources(6))
    paths_written.append(_compile_pyside_resources(2))


def _compile_uis(paths_written: typing.List[str]) -> None:
    ui_variants = {
        "pyside6": "pyside6-uic",
        "pyside2": "pyside2-uic",
    }
    for ui_subdir, uic_tool in ui_variants.items():
        ui_dir = f"cruiz/{ui_subdir}"
        if not os.path.isdir(ui_dir):
            os.makedirs(ui_dir)
        ui_basenames = [
            os.path.splitext(ui_file)[0]
            for ui_file in os.listdir("cruiz/resources")
            if os.path.splitext(ui_file)[1] == ".ui"
        ]
        for basename in ui_basenames:
            dst_file = f"{ui_dir}/{basename}.py"
            _run_command(
                [
                    uic_tool,
                    "-o",
                    dst_file,
                    f"cruiz/resources/{basename}.ui",
                ]
            )
            paths_written.append(dst_file)


def _capture_version(paths_written: typing.List[str]) -> None:
    version_path = "cruiz/RELEASE_VERSION.py"
    with open(version_path, "wt") as version_file:
        version_file.write(f'__version__ = "{get_version()}"')
    paths_written.append(version_path)


class _BuildPyCommand(setuptools.command.build_py.build_py):
    """
    Python build command, run during normal pip install, to generate data from
    resources.
    """

    def run(self) -> None:
        with _temp_resource_generation(_capture_version):
            with _temp_resource_generation(_compile_qrc):
                with _temp_resource_generation(_compile_uis):
                    super().run()


class _SDistCommand(setuptools.command.sdist.sdist):
    """
    Python source distribution command.
    """

    def run(self) -> None:
        with _temp_resource_generation(_capture_version):
            super().run()


requirements_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "requirements.txt"
)


with open(requirements_path) as requirements_file:
    requires = requirements_file.readlines()


readme_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "README.md"
)


with open(readme_path) as readme_file:
    readme = readme_file.read()


setup(
    cmdclass={
        "build_py": _BuildPyCommand,
        "sdist": _SDistCommand,
    },
    name="cruiz",
    version=get_version(),
    description="Conan recipe user interface",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Mark Final",
    author_email="markfinal@hotmail.com",
    url="https://github.com/markfinal/cruiz",
    packages=find_packages(),
    install_requires=requires,
    entry_points={
        "gui_scripts": [
            "cruiz = cruiz.__main__:main",
            "cruiz-pyside2 = cruiz.__main_pyside2__:main",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.7, <3.12",
    project_urls={
        'Documentation': 'https://cruiz.readthedocs.io/en/latest/',
        'GitHub': 'https://github.com/markfinal/cruiz',
    },
)



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing
import re

from cruiz.interop.searchrecipesparameters import SearchRecipesParameters

from .utils import worker
from .utils.message import Message, Success


def invoke(
    queue: multiprocessing.Queue[Message], params: SearchRecipesParameters
) -> None:
    """
    Equivalent to:
    'conan search -r <remote_name> [--case-sensitive] <pattern>'

    with optional extra processing when aliases are detected.

    SearchRecipesParameters has dynamic attributes.
    """
    with worker.ConanWorker(queue, params) as api:
        remote_name = params.remote_name  # type: ignore
        alias_aware = params.alias_aware  # type: ignore
        args = {
            "remote_name": remote_name,
            "case_sensitive": params.case_sensitive,  # type: ignore
        }
        try:
            # fill_revisions argument in Conan 1.18.0+
            # cannot use True as the value as it tries to inspect the local cache
            results = api.search_recipes(
                params.pattern,  # type: ignore
                fill_revisions=False,
                **args,
            )
        except TypeError:
            results = api.search_recipes(
                params.pattern,  # type: ignore
                **args,
            )

        if results["results"]:
            results_list = results["results"][0]["items"]
            package_references = [item["recipe"]["id"] for item in results_list]
        else:
            package_references = []
        filtered_results = []
        if alias_aware:
            predicate = re.compile(r'^\s*alias = "(.+)"$', re.MULTILINE)
            for ref in package_references:
                contents, _ = api.get_path(ref, None, None, remote_name)
                alias = None
                matches = predicate.findall(contents)
                if matches:
                    alias = matches[0]
                filtered_results.append((ref, alias))
        else:
            filtered_results = [(ref, None) for ref in package_references]

        queue.put(Success(filtered_results or None))



#!/usr/bin/env python3

"""
Settings context manager for recent recipes
"""

import typing

from qtpy import QtCore

from .basesettings import BaseSettings, CommonSettings


# TODO CommonSettings not all used
class RecentRecipeSettings(CommonSettings):
    """
    Settings for recent recipes
    """

    def uuids(self) -> typing.List[QtCore.QUuid]:
        """
        Get the list of UUIDs for the recent recipes.
        """
        result: typing.List[QtCore.QUuid] = []
        assert self.settings_reader
        settings = self.settings_reader.settings
        for i in range(self.settings_reader.count):
            settings.setArrayIndex(i)
            result.append(QtCore.QUuid(settings.value("UUID")))
        return result


class RecentRecipeSettingsReader:
    """
    Context manager for reading recent recipes from disk.
    """

    def __init__(self) -> None:
        self.group = ""
        self._array = "RecentRecipes"
        self.settings = BaseSettings.make_settings()
        self.count = 0

    def __enter__(self) -> RecentRecipeSettings:
        self.count = self.settings.beginReadArray(self._array)
        self._settings_object = RecentRecipeSettings()
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endArray()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class RecentRecipeSettingsWriter:  # TODO: should this derive from _WriterMixin?
    """
    Context manager for writing recent recipes to disk
    """

    def __init__(self) -> None:
        self._reader_for_writer = RecentRecipeSettingsReader()
        self._array = "RecentRecipes"

    def make_current(self, uuid: QtCore.QUuid) -> bool:
        """
        Make the specified recipe current
        """
        with self._reader_for_writer as settings:
            current_list = settings.uuids()
        uuid_exists = uuid in current_list
        # always make the current recipe the first in the list
        if uuid_exists:
            current_list.remove(uuid)
        current_list.insert(0, uuid)
        with BaseSettings.WriteArray(
            self._array, replace=True
        ) as settings:  # TODO: can this use the reader's QSettings?
            for i, uuid in enumerate(current_list):
                settings.setArrayIndex(i)
                settings.setValue("UUID", uuid.toString(QtCore.QUuid.WithoutBraces))
        return uuid_exists


class RecentRecipeSettingsDeleter:
    """
    Context manager for deleting a recent recipe
    """

    def __init__(self) -> None:
        self._reader_for_deleter = RecentRecipeSettingsReader()
        self._array = "RecentRecipes"

    def delete(self, uuid: QtCore.QUuid) -> None:
        """
        Delete the recent recipe represented by UUID
        """
        with self._reader_for_deleter as settings:
            current_list = settings.uuids()
        current_list = [u for u in current_list if u != uuid]
        BaseSettings.make_settings().remove(self._array)
        if current_list:
            with BaseSettings.WriteArray(self._array) as settings:
                for i, uuid in enumerate(current_list):
                    settings.setArrayIndex(i)
                    settings.setValue("UUID", uuid.toString(QtCore.QUuid.WithoutBraces))



#!/usr/bin/env python3

"""
Dialog for moving local caches
"""

import pathlib
import platform
import shutil
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

from cruiz.commands.context import ConanContext

from cruiz.settings.managers.namedlocalcache import (
    NamedLocalCacheSettings,
    NamedLocalCacheSettingsReader,
    NamedLocalCacheSettingsWriter,
)

if PYSIDE2:
    from cruiz.pyside2.local_cache_move import Ui_LocalCacheMove

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.local_cache_move import Ui_LocalCacheMove

    QAction = QtGui.QAction

from cruiz.widgets.util import search_for_dir_options


class MoveLocalCacheDialog(QtWidgets.QDialog):
    """
    Dialog for moving a non-default local cache.
    """

    def __init__(self, context: ConanContext, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        self._context = context
        self._ui = Ui_LocalCacheMove()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        with NamedLocalCacheSettingsReader(context.cache_name) as settings:
            self._ui.currentUserHome.setText(settings.home_dir.resolve())
            self._ui.currentUserHomeShort.setText(settings.short_home_dir.resolve())
        self._ui.buttonBox.accepted.connect(self.accept)
        self._ui.buttonBox.rejected.connect(self.reject)
        home_dir_browse_action = QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon), "", self
        )
        home_dir_browse_action.triggered.connect(self._home_dir_browse)
        self._ui.newUserHome.addAction(
            home_dir_browse_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.newUserHome.textChanged.connect(self._new_path_changed)
        self._ui.newUserHomeShort.textChanged.connect(self._new_path_changed)
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        if platform.system() == "Windows":
            pass
        else:
            self._ui.userHomeShortLabel.hide()
            self._ui.currentUserHomeShort.hide()
            self._ui.newUserHomeShort.hide()

    def _home_dir_browse(self) -> None:
        dir_found = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Select home directory to move to",
            self._ui.currentUserHome.text(),
            search_for_dir_options(),
        )
        if dir_found:
            self._ui.newUserHome.setText(dir_found)

    def _new_path_changed(self, path: str) -> None:
        # pylint: disable=unused-argument
        if platform.system() == "Windows":
            condition = bool(self._ui.newUserHome.text()) and bool(
                self._ui.newUserHomeShort.text()
            )
        else:
            condition = bool(self._ui.newUserHome.text()) or bool(
                self._ui.newUserHomeShort.text()
            )
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(condition)

    def accept(self) -> None:
        with NamedLocalCacheSettingsReader(self._context.cache_name) as settings:
            home_dir = settings.home_dir.resolve()
            short_home_dir = settings.short_home_dir.resolve()
        new_home_dir = self._ui.newUserHome.text().strip()
        if new_home_dir == home_dir:
            QtWidgets.QMessageBox.critical(
                self,
                "Conan local cache home directory",
                f"The selected home directory '{new_home_dir}' is unchanged. "
                "Please choose another.",
            )
            return
        qdir = QtCore.QDir(new_home_dir)
        if qdir.exists():
            QtWidgets.QMessageBox.critical(
                self,
                "Conan local cache home directory",
                f"The selected home directory '{new_home_dir}' already exists. "
                "Please choose another.",
            )
            return
        if not qdir.isEmpty():
            QtWidgets.QMessageBox.critical(
                self,
                "Conan local cache home directory",
                f"The selected home directory '{new_home_dir}' is not empty. "
                "Please choose another.",
            )
            return
        if short_home_dir:
            new_short_home_dir = self._ui.newUserHomeShort.text().strip()
            if new_short_home_dir == short_home_dir:
                QtWidgets.QMessageBox.critical(
                    self,
                    "Conan local cache short home directory",
                    f"The selected short home directory '{new_short_home_dir}' "
                    "is unchanged. Please choose another.",
                )
                return
            qdir = QtCore.QDir(new_short_home_dir)
            if qdir.exists():
                QtWidgets.QMessageBox.critical(
                    self,
                    "Conan local cache short home directory",
                    f"The selected short home directory '{new_short_home_dir}' already"
                    " exists. Please choose another.",
                )
                return
            if not qdir.isEmpty():
                QtWidgets.QMessageBox.critical(
                    self,
                    "Conan local cache short home directory",
                    f"The selected short home directory '{new_short_home_dir}' is "
                    "not empty. Please choose another.",
                )
                return
        else:
            new_short_home_dir = None
        self._move(settings, home_dir, short_home_dir, new_home_dir, new_short_home_dir)
        super().accept()

    def _move(
        self,
        settings: NamedLocalCacheSettings,
        old_home_dir: str,
        old_short_home_dir: str,
        new_home_dir: str,
        new_short_home_dir: typing.Optional[str],
    ) -> None:
        assert old_home_dir  # since it's non-default
        # update settings
        settings.home_dir = new_home_dir  # type: ignore
        settings.short_home_dir = new_short_home_dir  # type: ignore
        NamedLocalCacheSettingsWriter(self._context.cache_name).sync(settings)
        # pylint: disable=broad-except
        try:
            # move <old>/.conan to <new>/.conan
            old_conan_dir = pathlib.Path(old_home_dir) / ".conan"
            new_conan_dir = pathlib.Path(new_home_dir) / ".conan"
            shutil.move(str(old_conan_dir), str(new_conan_dir))
        except Exception as exception:
            QtWidgets.QMessageBox.critical(
                self,
                "Conan move local cache",
                f"Failed to move the local cache home dir because: {exception}.",
            )
            # roll back previous steps
            settings.home_dir = old_home_dir  # type: ignore
            settings.short_home_dir = old_short_home_dir  # type: ignore
            NamedLocalCacheSettingsWriter(self._context.cache_name).sync(settings)
            return
        else:
            if old_short_home_dir:
                try:
                    assert new_short_home_dir
                    shutil.move(old_short_home_dir, new_short_home_dir)
                except Exception as exception:
                    QtWidgets.QMessageBox.critical(
                        self,
                        "Conan move local cache",
                        "Failed to move the local cache short home dir because: "
                        f"{exception}.",
                    )
                    # roll back previous steps
                    shutil.move(new_home_dir, str(old_conan_dir))
                    settings.home_dir = old_home_dir  # type: ignore
                    settings.short_home_dir = old_short_home_dir  # type: ignore
                    NamedLocalCacheSettingsWriter(self._context.cache_name).sync(
                        settings
                    )
                    return
        qdir = QtCore.QDir(old_home_dir)
        if qdir.isEmpty():
            qdir.removeRecursively()



#!/usr/bin/env python3

"""
Logging details
"""

import typing

from qtpy import QtCore, QtWidgets

from .guardedlisttoflush import GuardedListToFlush


class LogDetails(QtCore.QObject):
    """
    Representation of how and where to perform logging during commands
    """

    logging = QtCore.Signal()

    def __init__(
        self,
        output: QtWidgets.QPlainTextEdit,
        error: typing.Optional[QtWidgets.QPlainTextEdit],
        combined: bool,
        batched: bool,
        conan_log: typing.Optional[QtWidgets.QPlainTextEdit],
    ) -> None:
        super().__init__()
        self.output = output
        self.error = output if combined else error
        self._combined = combined
        if combined:
            self._stdout_list = GuardedListToFlush() if batched else None
            self._stderr_list = self._stdout_list
        else:
            self._stdout_list = GuardedListToFlush() if batched else None
            self._stderr_list = GuardedListToFlush() if batched else None
        self._conan_log = conan_log

    def start(self) -> None:
        """
        Start logging
        """
        if self._stdout_list:
            self._stdout_list.start(1000, self.output)
        if self._stderr_list and self._stderr_list != self._stdout_list:
            err_widget = self.output if self._combined else self.error
            assert err_widget
            self._stderr_list.start(1000, err_widget)

    def stop(self) -> None:
        """
        Stop logging
        """
        if self._stdout_list:
            self._stdout_list.stop()
        if self._stderr_list and self._stderr_list != self._stdout_list:
            self._stderr_list.stop()

    def stdout(self, text: str) -> None:
        """
        Append text to stdout
        """
        if self._stdout_list:
            self._stdout_list.append(text)
        else:
            self.output.appendHtml(text)
        self.logging.emit()

    def stderr(self, text: str) -> None:
        """
        Append text to stderr
        """
        if self._stderr_list:
            self._stderr_list.append(text)
        else:
            assert self.error
            self.error.appendHtml(text)
        self.logging.emit()

    def conan_log(self, text: str) -> None:
        """
        Append a Conan log message
        """
        if self._conan_log:
            self._conan_log.appendPlainText(text)



#!/usr/bin/env python3

"""
Representation on the host of an instance of a Conan or Conan-related
command running in a child process
"""

import logging
import multiprocessing
import os
import signal
import sys
import threading
import typing

import psutil

from qtpy import QtCore, QtWidgets

from cruiz.settings.managers.generalpreferences import GeneralSettingsReader
from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.searchrecipesparameters import SearchRecipesParameters
from cruiz.interop.reciperevisionsparameters import RecipeRevisionsParameters
from cruiz.interop.packageidparameters import PackageIdParameters
from cruiz.interop.packagerevisionsparameters import PackageRevisionsParameters
from cruiz.interop.packagebinaryparameters import PackageBinaryParameters

from .messagereplyprocessor import MessageReplyProcessor

from .logdetails import LogDetails


logger = logging.getLogger(__name__)


class ConanInvocation(QtCore.QObject):
    """
    Wrapper around invoking a conan child process
    """

    # pylint: disable=too-many-instance-attributes

    completed = QtCore.Signal(object, Exception)
    finished = QtCore.Signal()
    _begin_processing = QtCore.Signal()

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(self, log_details: LogDetails) -> None:
        logger.debug("+=%d", id(self))
        super().__init__()  # note that parent is None
        self._log_details = log_details
        self._mp_context = multiprocessing.get_context("spawn")
        self._process_queue = self._mp_context.Queue()
        self._thread = QtCore.QThread()
        self._queue_processor = MessageReplyProcessor(self._process_queue)
        self._queue_processor.moveToThread(self._thread)
        self._begin_processing.connect(self._queue_processor.process)
        self._queue_processor.completed.connect(self.completed)
        # self._thread.finished not guaranteed to be delivered in a
        # qApp quitting scenario
        self._thread.finished.connect(self._thread.deleteLater)
        self._thread.finished.connect(self.finished)
        self._process: typing.Optional[multiprocessing.context.SpawnProcess] = None
        self._last_command_running: bool = False  # TODO: remove this
        self._cleanup_thread: typing.Optional[threading.Thread] = None

        self._thread.start()
        self._begin_processing.emit()

    def close(self) -> None:
        """
        Tidy up any resources on the context that need closing
        """
        if self._process:
            self._process.join()
            self._process.close()
        self._process = None
        self._queue_processor.stop()
        self._thread.wait()

    def _disconnect_signal(self, result: typing.Any, exception: typing.Any) -> None:
        # pylint: disable=unused-argument
        self._queue_processor.completed.disconnect()
        self._last_command_running = False

    def _critical_failure(self, message: str) -> None:
        QtWidgets.QMessageBox.critical(
            None, "System failure", message  # type: ignore[call-overload]
        )
        sys.exit(1)

    def invoke(
        self,
        parameters: typing.Union[
            CommandParameters,
            SearchRecipesParameters,
            RecipeRevisionsParameters,
            PackageIdParameters,
            PackageRevisionsParameters,
            PackageBinaryParameters,
        ],
        log_details: LogDetails,
        continuation: typing.Optional[typing.Callable[[typing.Any, typing.Any], None]],
    ) -> None:
        """
        Invoke a command, with optional continuation
        """
        assert self._process is None

        if continuation:
            self._queue_processor.completed.connect(continuation)
        self._queue_processor.completed.connect(self._disconnect_signal)

        self._queue_processor.stdout_message.connect(log_details.stdout)
        self._queue_processor.stderr_message.connect(log_details.stderr)
        self._queue_processor.conan_log_message.connect(log_details.conan_log)
        self._queue_processor.critical_failure.connect(self._critical_failure)
        with GeneralSettingsReader() as settings:
            clear_panes = settings.clear_panes.resolve()
        if clear_panes and isinstance(log_details, LogDetails):
            log_details.output.clear()
            if log_details.error:
                log_details.error.clear()

        process = self._mp_context.Process(
            target=parameters.worker,
            args=(self._process_queue, parameters),
            daemon=False,
        )
        process.start()
        logger.debug(
            "cruiz (pid=%i) started child process (pid=%i) for %s",
            os.getpid(),
            process.pid,
            parameters.worker.__module__,
        )
        self._process = process

    def cancel(self) -> None:
        """
        Cancel the current running worker thread.
        """
        if not self._process:
            return
        if not self._process.is_alive():
            return

        pid = self._process.pid
        current_process_psutil = psutil.Process(pid)
        children = current_process_psutil.children(recursive=True)
        self._process.terminate()  # orphans children

        # spawn a daemonic thread to do the child process cleanup
        # doing it in-place was causing shutdown issues on Windows at least
        self._cleanup_thread = threading.Thread(
            target=ConanInvocation._cleanup, args=(pid, children)
        )
        self._cleanup_thread.daemon = True
        self._cleanup_thread.start()

    @staticmethod
    def _cleanup(pid: int, children: typing.List[psutil.Process]) -> None:
        logger.debug("Worker process: %i cancelled", pid)
        logger.debug("Had %i child processes", len(children))
        for i, child in enumerate(children):
            # pylint: disable=broad-except
            try:
                with child.oneshot():
                    logger.debug(
                        "\t%d: %s - %s - %s",
                        i,
                        child.name(),
                        child.cmdline(),
                        child.status(),
                    )
                child.send_signal(signal.SIGTERM)
            except Exception as exception:
                logger.debug("\t%d: %s", i, str(exception))
        psutil.wait_procs(children)



#!/usr/bin/env python3

"""
Useful constants
"""

from enum import Enum, auto


class BuildFeatureConstants(Enum):
    """
    Constants representing build features
    """

    CMAKEFINDDEBUGMODE = auto()
    CMAKEVERBOSEMODE = auto()
    CCACHEEXECUTABLE = auto()
    CCACHEAUTOTOOLSCONFIGARGS = auto()
    SCCACHEEXECUTABLE = auto()
    SCCACHEAUTOTOOLSCONFIGARGS = auto()
    BUILDCACHEEXECUTABLE = auto()
    BUILDCACHEAUTOTOOLSCONFIGARGS = auto()


class CompilerCacheTypes(Enum):
    """
    Compiler cache types
    """

    NONE = None
    CCACHE = "ccache"
    SCCACHE = "sccache"
    BUILDCACHE = "buildcache"


DEFAULT_CACHE_NAME = "Default"



#!/usr/bin/env python3

"""
Conan dependency view
"""

import typing

from qtpy import QtWidgets, QtSvg, PYSIDE2

if PYSIDE2:
    QGraphicsSvgItem = QtSvg.QGraphicsSvgItem
else:
    # module not present in qtpy at the time of writing
    from PySide6 import QtSvgWidgets

    QGraphicsSvgItem = QtSvgWidgets.QGraphicsSvgItem

from cruiz.interop.dependencygraph import DependencyGraph
from cruiz.svggraph import DependenciesToDigraph, DigraphToSVG, SVGScene


class DependencyView(QtWidgets.QGraphicsView):
    def clear(self) -> None:
        self.setScene(None)  # type: ignore

    def visualise(self, depgraph: DependencyGraph, rankdir: int) -> None:
        if rankdir == 0:
            rank_dir = "LR"
        elif rankdir == 1:
            rank_dir = "TB"
        digraph = DependenciesToDigraph(depgraph, rank_dir)
        svg = DigraphToSVG(digraph)
        self._scene = SVGScene(QtSvg.QSvgRenderer(svg.svg))
        self.setScene(self._scene)


class InverseDependencyViewDialog(QtWidgets.QDialog):
    def __init__(
        self,
        depgraph: DependencyGraph,
        parent: typing.Optional[QtWidgets.QWidget] = None,
    ) -> None:
        super().__init__(parent)
        self.setWindowTitle(f"What uses {depgraph.root.name}")

        # TODO: should this be in cruiz.svggraph to isolate the SVG code?
        digraph = DependenciesToDigraph(depgraph, "LR", flipped_edges=True)
        svg = DigraphToSVG(digraph)
        self._renderer = QtSvg.QSvgRenderer(svg.svg)
        item = QGraphicsSvgItem()
        item.setSharedRenderer(self._renderer)
        self._scene = QtWidgets.QGraphicsScene()
        self._scene.addItem(item)
        view = QtWidgets.QGraphicsView(self._scene)

        QtWidgets.QVBoxLayout(self).addWidget(view)



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.packagenode import PackageNode
from cruiz.interop.dependencygraph import DependencyGraph

from .utils import worker
from .utils.formatoptions import format_options
from .utils.message import Message, Success


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan lock create'
    """
    with worker.ConanWorker(queue, params) as api:
        # code derived from "conan lock create" command,
        # but aborting before writing a lock file to disk
        from conans.client.profile_loader import profile_from_args
        from conans.model.ref import ConanFileReference
        from conans.model.graph_info import GraphInfo
        from conans.model.env_info import EnvValues
        from conans.client.recorder.action_recorder import ActionRecorder
        from conans.client.graph.graph import (
            RECIPE_CONSUMER,
            RECIPE_VIRTUAL,
        )
        from conans.model.ref import PackageReference

        assert params.profile, "Profile is needed for creating a lock file"

        try:
            phost = profile_from_args(
                [params.profile],
                None,
                format_options(params.options),
                None,
                None,
                params.cwd,
                api.app.cache,
                build_profile=False,
            )
        except TypeError:
            # no conf and build_profile parameters in older Conans
            phost = profile_from_args(
                [params.profile],
                None,
                format_options(params.options),
                None,
                params.cwd,
                api.app.cache,
            )

        phost.process_settings(api.app.cache)

        root_ref = ConanFileReference(
            params.name, params.version, params.user, params.channel, validate=False
        )
        graph_info = GraphInfo(
            profile_host=phost, profile_build=None, root_ref=root_ref
        )

        remotes = api.app.load_remotes()
        recorder = ActionRecorder()

        deps_graph = api.app.graph_manager.load_graph(
            str(params.recipe_path),
            None,
            graph_info,
            None,
            None,
            None,
            remotes,
            recorder,
        )

        # create nodes (derived from conan.graph.printer.print_graph)
        build_time_nodes = deps_graph.build_time_nodes()
        nodes = {}
        for node in sorted(deps_graph.nodes):
            # ensure some attributes, otherwise ConanInfo.dumps() fails
            try:
                if not hasattr(node.conanfile.original_info, "recipe_hash"):
                    node.conanfile.original_info.recipe_hash = None
                if not hasattr(node.conanfile.original_info, "env_values"):
                    node.conanfile.original_info.env_values = EnvValues()
                info = node.conanfile.original_info.dumps()
            except AttributeError:
                # node.conanfile.original_info is only available from Conan 1.47.0+
                info = None

            if node.recipe in (RECIPE_CONSUMER, RECIPE_VIRTUAL):
                new_node = PackageNode(
                    node.name,
                    str(node.ref),
                    node.package_id,
                    node.ref.revision,
                    node.conanfile.short_paths,
                    info,
                    True,
                )
                nodes[node] = new_node
                continue
            # ignoring the return value from...
            PackageReference(node.ref, node.package_id, node.ref.revision)
            is_runtime = node not in build_time_nodes
            new_node = PackageNode(
                node.name,
                str(node.ref),
                node.package_id,
                node.ref.revision,
                node.conanfile.short_paths,
                info,
                is_runtime,
            )
            nodes[node] = new_node

        # connect nodes (two-way)
        for node, interop_node in nodes.items():
            for dep in node.dependencies:
                interop_node.children.append(nodes[dep.dst])
                nodes[dep.dst].parents.append(interop_node)

        new_graph = DependencyGraph(list(nodes.values()), new_node)

        queue.put(Success(new_graph))



#!/usr/bin/env python3

"""
Recipe behaviour toolbar
"""

import os
import pathlib

from qtpy import QtCore, QtWidgets, PYSIDE2

if PYSIDE2:
    from cruiz.pyside2.recipe_profile_frame import Ui_profileFrame
    from cruiz.pyside2.recipe_cpucores_frame import Ui_cpuCoresFrame
else:
    from cruiz.pyside6.recipe_profile_frame import Ui_profileFrame
    from cruiz.pyside6.recipe_cpucores_frame import Ui_cpuCoresFrame

from cruiz.widgets.util import BlockSignals

from cruiz.settings.managers.recipe import (
    RecipeSettings,
    RecipeSettingsReader,
    RecipeSettingsWriter,
)
from cruiz.settings.managers.namedlocalcache import NamedLocalCacheSettingsReader

from cruiz.recipe.recipe import Recipe


class _ProfileFrame(QtWidgets.QFrame):
    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self._ui = Ui_profileFrame()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._ui.profileCombo.currentTextChanged.connect(self._changed_profile)

    @property
    def _recipe(self) -> Recipe:
        toolbar = self.parent()
        recipe_widget = toolbar.parent()
        recipe = recipe_widget.recipe
        return recipe

    def _refresh_tooltip(self) -> None:
        cache_name = self._recipe.context.cache_name
        with NamedLocalCacheSettingsReader(cache_name) as settings:
            home_dir = settings.home_dir.resolve()
        # using normcase for Windows path compares
        home_dir = os.path.normcase(home_dir) if home_dir else str(pathlib.Path.home())
        profile_dirs = "\n  ".join(
            [str(os.path.normcase(p)) for p in self._recipe.context.all_profile_dirs()]
        )
        parameterised_profile_dirs = profile_dirs.replace(home_dir, "<CONAN_USER_HOME>")
        tooltip = (
            "Profiles listed here are from the following directories:\n\n"
            f"{profile_dirs}\n\n"
            "which can be read as this parameterisation\n\n"
            f"{parameterised_profile_dirs}\n\n"
            "where <CONAN_USER_HOME> is the base directory "
            "for this recipe's associated local cache, called "
            f"'{cache_name}'"
        )
        self._ui.profileCombo.setToolTip(tooltip)
        self._ui.profileLabel.setToolTip(tooltip)

    def refresh_content(self, recipe: Recipe) -> None:
        """
        Refresh the contents.
        """
        profile_list = recipe.context.get_list_of_profiles()
        with RecipeSettingsReader.from_recipe(recipe) as settings:
            current_profile = settings.profile.resolve()
        with BlockSignals(self._ui.profileCombo) as blocked_widget:
            blocked_widget.clear()
            for profile_path, _ in profile_list:
                blocked_widget.addItem(str(profile_path))

            index = blocked_widget.findText(current_profile)
            if index >= 0:
                blocked_widget.setCurrentIndex(index)
            else:
                # this means that the profile in settings cannot be found in the
                # local cache profiles
                blocked_widget.setCurrentIndex(0)
                # sync this choice back to the settings
                updated_profile_settings = RecipeSettings()
                updated_profile_settings.profile = str(
                    profile_list[0][0]
                )  # type: ignore
                RecipeSettingsWriter().from_recipe(recipe).sync(
                    updated_profile_settings
                )
        self._refresh_tooltip()

    def _changed_profile(self, text: str) -> None:
        recipe = self._recipe
        settings = RecipeSettings.from_recipe(recipe)
        settings.profile = text  # type: ignore
        # note: has to be from recipe, as the profile setter needs
        # the ConanContext to check the default profile name
        RecipeSettingsWriter.from_recipe(recipe).sync(settings)
        self.parent().profile_changed.emit(text)


class _CPUCoresFrame(QtWidgets.QFrame):
    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self._ui = Ui_cpuCoresFrame()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._ui.cpuCoresSpin.valueChanged.connect(self._changed_num_cores)

    @property
    def _recipe(self) -> Recipe:
        toolbar = self.parent()
        recipe_widget = toolbar.parent()
        recipe = recipe_widget.recipe
        return recipe

    def refresh_content(self, recipe: Recipe) -> None:
        """
        Refresh the contents.
        """
        with RecipeSettingsReader.from_recipe(recipe) as settings:
            num_cores = settings.num_cpu_cores
        if num_cores.value is None:
            self._ui.cpuCoresSpin.setStyleSheet("color: green;")
        elif num_cores.value > num_cores.fallback:
            self._ui.cpuCoresSpin.setStyleSheet("color: orange;")
        elif num_cores.value < num_cores.fallback:
            self._ui.cpuCoresSpin.setStyleSheet("color: red;")
        with BlockSignals(self._ui.cpuCoresSpin) as blocked_widget:
            blocked_widget.setValue(num_cores.resolve())

    def _changed_num_cores(self, value: int) -> None:
        recipe = self._recipe
        settings = RecipeSettings.from_recipe(recipe)
        settings.num_cpu_cores = value  # type: ignore
        RecipeSettingsWriter.from_recipe(recipe).sync(settings)
        self.refresh_content(recipe)


class RecipeBehaviourToolbar(QtWidgets.QToolBar):
    """
    QToolBar representing tne behaviour of a recipe, profile and cpu cores
    """

    profile_changed = QtCore.Signal(str)

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self._profile = _ProfileFrame(self)
        self._cpu_cores = _CPUCoresFrame(self)
        self.addWidget(self._profile)
        self.addSeparator()
        self.addWidget(self._cpu_cores)

    @property
    def _recipe(self) -> Recipe:
        recipe_widget = self.parent()
        recipe = recipe_widget.recipe
        return recipe

    def refresh_content(self) -> None:
        """
        Refresh the content of the toolbar.
        """
        recipe = self._recipe
        self._profile.refresh_content(recipe)
        self._cpu_cores.refresh_content(recipe)



#!/usr/bin/env python3

"""
Wizard page for selecting a local cache
"""

import typing

from qtpy import QtWidgets

from cruiz.widgets.util import BlockSignals
from cruiz.settings.managers.namedlocalcache import AllNamedLocalCacheSettingsReader
from cruiz.settings.managers.recipe import RecipeSettingsReader
from cruiz.manage_local_cache import ManageLocalCachesDialog


class LoadRecipeLocalCachePage(QtWidgets.QWizardPage):
    """
    Wizard page for selecting the local cache to associate with the recipe instance.
    """

    @property
    def _ui(self) -> typing.Any:
        return self.wizard().ui

    def nextId(self) -> int:
        return 3

    def initializePage(self) -> None:
        self.registerField(
            "local_cache*", self._ui.local_cache_name, "currentText"  # type: ignore
        )
        self._ui.local_cache_name.currentIndexChanged.connect(
            self._on_local_cache_changed
        )
        self._ui.manage_caches.clicked.connect(self._on_manage_caches)
        self._populate_caches()
        if self.wizard().original_recipe:
            with RecipeSettingsReader.from_recipe(
                self.wizard().original_recipe
            ) as settings:
                original_cache = settings.local_cache_name.resolve()
            self._ui.local_cache_name.setCurrentText(original_cache)
        super().initializePage()

    def cleanupPage(self) -> None:
        self._ui.local_cache_name.currentIndexChanged.disconnect()
        return super().cleanupPage()

    def _populate_caches(self) -> None:
        with BlockSignals(self._ui.local_cache_name) as blocked_widget:
            blocked_widget.clear()
            with AllNamedLocalCacheSettingsReader() as names:
                blocked_widget.addItems(names)
            blocked_widget.setCurrentIndex(-1)

    def _on_local_cache_changed(self, index: int) -> None:
        # pylint: disable=unused-argument
        self.completeChanged.emit()

    def _on_manage_caches(self) -> None:
        ManageLocalCachesDialog(self, None).exec_()
        self._populate_caches()



#!/usr/bin/env python3

"""
Settings context manager for CMake preferences
"""

import typing

from .basesettings import (
    BaseSettings,
    StringSetting,
    ComparableCommonSettings,
    SettingMeta,
)
from .writermixin import _WriterMixin
from .valueclasses import ScalarValue


class CMakeSettings(ComparableCommonSettings):
    """
    Representation of CMake settings
    """

    def __init__(self) -> None:
        self._property_meta = {
            "bin_directory": SettingMeta("BinDir", StringSetting, None, ScalarValue),
        }

    @property
    def bin_directory(self) -> StringSetting:
        """
        Get the CMake bin directory
        """
        return self._get_value_via_meta()

    @bin_directory.setter
    def bin_directory(self, value: str) -> None:
        self._set_value_via_meta(value)


class CMakeSettingsReader:
    """
    Context manager for reading CMake settings from disk
    """

    def __init__(self) -> None:
        self.group = "Thirdparty/CMake"
        self.settings = BaseSettings.make_settings()

    def __enter__(self) -> CMakeSettings:
        self.settings.beginGroup(self.group)
        settings = CMakeSettings()
        settings.settings_reader = self
        return settings

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endGroup()
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class CMakeSettingsWriter(_WriterMixin):
    """
    Utility class for writing changed CMake settings to disk
    """

    def __init__(self) -> None:
        self._reader_for_writer = CMakeSettingsReader()



#!/usr/bin/env python3

"""
Qt logger for Conan
"""

from __future__ import annotations

import logging
import multiprocessing
import typing

import conans.util.log

import cruiz.workers.utils.message


class _Singleton(type):
    """
    Base class for all singletons
    """

    _instances: typing.Dict[typing.Any, typing.Any] = {}

    def __call__(cls, *args: typing.Any, **kwargs: typing.Any) -> _Singleton:
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class QtLogger(logging.Handler, metaclass=_Singleton):
    """
    A logger handler that can be used by Qt to emit signals.
    Signals are delegated to another class that derives from QObject
    This is a singleton.
    """

    def __init__(self) -> None:
        super().__init__()
        formatter = conans.util.log.MultiLineFormatter(
            "%(levelname)-6s:%(filename)-15s[%(lineno)d]: %(message)s [%(asctime)s]"
        )
        self.setFormatter(formatter)
        self._queue: typing.Optional[
            multiprocessing.Queue[cruiz.workers.utils.message.Message]
        ] = None

    def set_queue(
        self, queue: multiprocessing.Queue[cruiz.workers.utils.message.Message]
    ) -> None:
        """
        Set the multiprocessing queue to send messages with.
        """
        self._queue = queue

    def emit(self, record: typing.Any) -> None:
        assert self._queue
        self._queue.put(
            cruiz.workers.utils.message.ConanLogMessage(self.format(record))
        )



#!/usr/bin/env python3

"""
Util to set the process' environment
"""

import multiprocessing
import os
import typing


def clear_conan_env() -> None:
    """
    cruiz needs to control all Conan environments, so ignore everything from
    the calling environment
    """
    external_conan_envvar_names = [
        key for key, _ in os.environ.items() if key.startswith("CONAN_")
    ]
    if external_conan_envvar_names:
        multiprocessing.get_logger().debug(
            "Removing the following environment variables from cruiz's process:"
        )
        for key in external_conan_envvar_names:
            multiprocessing.get_logger().debug("\t%s", key)
            os.environ.pop(key)


def set_env(
    additions: typing.Dict[str, typing.Any], removals: typing.List[str]
) -> None:
    """
    Set the environment given key-value pairs to add, and keys to remove.
    If a key is added, and already exists in the public environment map,
    then it is prepended with a path separator.
    """
    for key, value in additions.items():
        if value is None:
            continue
        if not isinstance(value, str):
            value = str(value)
        if key in os.environ:
            if value not in os.environ[key]:
                os.environ[key] = value + os.pathsep + os.environ[key]
        else:
            os.environ[key] = value
    for key in removals:
        if key in os.environ:
            os.environ.pop(key)



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing
import typing

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan export-pkg'
    """
    with worker.ConanWorker(queue, params) as api:
        assert params.recipe_path
        assert params.cwd
        assert params.name
        args: typing.Dict[str, typing.Union[bool, str]] = {}
        if params.version:
            args["version"] = params.version
        if params.user:
            args["user"] = params.user
        if params.install_folder:
            args["install_folder"] = str(params.install_folder)
        if params.source_folder:
            args["source_folder"] = str(params.source_folder)
        if params.build_folder:
            args["build_folder"] = str(params.build_folder)
        if params.package_folder:
            args["package_folder"] = str(params.package_folder)
        if params.force:
            args["force"] = params.force

        result = api.export_pkg(
            str(params.recipe_path),
            params.name,
            params.channel or None,
            **args,
        )

        worker.replace_conan_version_struct_with_string(result)

        queue.put(Success(result))



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan build'
    """
    with worker.ConanWorker(queue, params) as api:
        assert params.recipe_path
        assert params.cwd
        args = {}
        if params.install_folder:
            args["install_folder"] = params.install_folder
        if params.source_folder:
            args["source_folder"] = params.source_folder
        if params.build_folder:
            args["build_folder"] = params.build_folder
        if params.package_folder:
            args["package_folder"] = params.package_folder

        result = api.build(str(params.recipe_path), **args)
        queue.put(Success(result))



#!/usr/bin/env python3

"""
Any monkeypatching required
"""

import logging
import os
import typing

import conans

from cruiz.constants import BuildFeatureConstants


def _monkey_patch_logger() -> None:

    # Monkey patched version of conans.util.log.configure_logger
    # that does not always remove handlers, keeps the qt handler
    def qt_configure_logger(
        logging_level: int = logging.CRITICAL, logging_file: typing.Optional[str] = None
    ) -> logging.Logger:
        # pylint: disable=unused-argument
        # pylint: disable=import-outside-toplevel
        import cruiz.workers.utils.qtlogger

        logger = logging.getLogger("conans")
        qt_logger = cruiz.workers.utils.qtlogger.QtLogger()
        if qt_logger in logger.handlers:
            return logger
        for hand in logger.handlers:
            logger.removeHandler(hand)
        logger.addHandler(qt_logger)
        logger.setLevel(logging_level)
        logger.propagate = False
        return logger

    conans.util.log.configure_logger = qt_configure_logger


try:
    # conan < 1.34.0
    original_cmake_build_helper_configure = (
        conans.client.build.cmake.CMakeBuildHelper.configure
    )
except AttributeError:
    try:
        original_cmake_build_helper_configure = (
            conans.client.build.cmake.CMake.configure
        )
    except AttributeError:
        print("CONAN 2: Fix up CMake configure monkey patch")


def _monkey_patch_cmake_helper() -> None:
    def new_configure(
        self: typing.Any,
        args: typing.Any = None,
        defs: typing.Any = None,
        source_dir: typing.Any = None,
        build_dir: typing.Any = None,
        source_folder: typing.Any = None,
        build_folder: typing.Any = None,
        cache_build_folder: typing.Any = None,
        pkg_config_paths: typing.Any = None,
    ) -> None:
        # CMake find debug support
        if str(BuildFeatureConstants.CMAKEFINDDEBUGMODE) in os.environ:
            defs = defs or {}
            defs["CMAKE_FIND_DEBUG_MODE"] = "TRUE"

        # CMake verbose support
        if str(BuildFeatureConstants.CMAKEVERBOSEMODE) in os.environ:
            defs = defs or {}
            defs["CMAKE_VERBOSE_MAKEFILE"] = "TRUE"

        # CMake compiler cache support
        cache_executable = None
        if str(BuildFeatureConstants.CCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[str(BuildFeatureConstants.CCACHEEXECUTABLE)]
        if str(BuildFeatureConstants.SCCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[str(BuildFeatureConstants.SCCACHEEXECUTABLE)]
        if str(BuildFeatureConstants.BUILDCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[
                str(BuildFeatureConstants.BUILDCACHEEXECUTABLE)
            ]
        if cache_executable:
            defs = defs or {}
            defs["CMAKE_C_COMPILER_LAUNCHER"] = cache_executable
            defs["CMAKE_CXX_COMPILER_LAUNCHER"] = cache_executable

        # now execute the old function
        original_cmake_build_helper_configure(
            self,
            args,
            defs,
            source_dir,
            build_dir,
            source_folder,
            build_folder,
            cache_build_folder,
            pkg_config_paths,
        )

    try:
        # conan < 1.34.0
        conans.client.build.cmake.CMakeBuildHelper.configure = new_configure
    except AttributeError:
        try:
            conans.client.build.cmake.CMake.configure = new_configure
        except AttributeError:
            print("CONAN 2: Fix up CMake monkey patch")


try:
    original_autotools_build_helper_configure = (
        conans.client.build.autotools_environment.AutoToolsBuildEnvironment.configure
    )

    original_autotools_build_helper_make = (
        conans.client.build.autotools_environment.AutoToolsBuildEnvironment.make
    )
except AttributeError:
    print("CONAN 2: Fix up Autotools monkey patch")


def _monkey_patch_autotools_helper() -> None:
    def new_configure(
        self: typing.Any,
        configure_dir: typing.Any = None,
        args: typing.Any = None,
        build: typing.Any = None,
        host: typing.Any = None,
        target: typing.Any = None,
        pkg_config_paths: typing.Any = None,
        vars: typing.Any = None,
        use_default_install_dirs: typing.Any = True,
    ) -> None:
        # if using autotools, GCC is almost implied
        # on macOSX, gcc/g++ do exist and do report being apple clang
        cache_executable = None
        if str(BuildFeatureConstants.CCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[str(BuildFeatureConstants.CCACHEEXECUTABLE)]
            if str(BuildFeatureConstants.CCACHEAUTOTOOLSCONFIGARGS) in os.environ:
                args = args or []
                args.extend(
                    os.environ[
                        str(BuildFeatureConstants.CCACHEAUTOTOOLSCONFIGARGS)
                    ].split(" ")
                )
        if str(BuildFeatureConstants.SCCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[str(BuildFeatureConstants.SCCACHEEXECUTABLE)]
            if str(BuildFeatureConstants.SCCACHEAUTOTOOLSCONFIGARGS) in os.environ:
                args = args or []
                args.extend(
                    os.environ[
                        str(BuildFeatureConstants.SCCACHEAUTOTOOLSCONFIGARGS)
                    ].split(" ")
                )
        if str(BuildFeatureConstants.BUILDCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[
                str(BuildFeatureConstants.BUILDCACHEEXECUTABLE)
            ]
            if str(BuildFeatureConstants.BUILDCACHEAUTOTOOLSCONFIGARGS) in os.environ:
                args = args or []
                args.extend(
                    os.environ[
                        str(BuildFeatureConstants.BUILDCACHEAUTOTOOLSCONFIGARGS)
                    ].split(" ")
                )
        if cache_executable:
            vars = vars or {}
            vars["CC"] = f"{cache_executable} gcc"
            vars["CXX"] = f"{cache_executable} g++"
            vars["PATH"] = [os.path.dirname(cache_executable)]

        # now execute the old function
        original_autotools_build_helper_configure(
            self,
            configure_dir,
            args,
            build,
            host,
            target,
            pkg_config_paths,
            vars,
            use_default_install_dirs,
        )

    try:
        conans.client.build.autotools_environment.AutoToolsBuildEnvironment.configure = (  # noqa: E501
            new_configure
        )
    except AttributeError:
        print("CONAN 2: Fix up Autotools monkey patch")

    def new_make(
        self: typing.Any,
        args: str = "",
        make_program: typing.Optional[str] = None,
        target: typing.Optional[str] = None,
        vars: typing.Any = None,
    ) -> None:
        cache_executable = None
        if str(BuildFeatureConstants.CCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[str(BuildFeatureConstants.CCACHEEXECUTABLE)]
        if str(BuildFeatureConstants.SCCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[str(BuildFeatureConstants.SCCACHEEXECUTABLE)]
        if str(BuildFeatureConstants.BUILDCACHEEXECUTABLE) in os.environ:
            cache_executable = os.environ[
                str(BuildFeatureConstants.BUILDCACHEEXECUTABLE)
            ]
        if cache_executable:
            vars = vars or {}
            vars["CC"] = f"{cache_executable} gcc"
            vars["CXX"] = f"{cache_executable} g++"
            vars["PATH"] = [os.path.dirname(cache_executable)]

        # now execute the old function
        original_autotools_build_helper_make(self, args, make_program, target, vars)

    try:
        conans.client.build.autotools_environment.AutoToolsBuildEnvironment.make = (
            new_make
        )
    except AttributeError:
        print("CONAN 2: Fix up Autotools monkey patch")


def _do_monkey_patching() -> None:
    _monkey_patch_logger()
    _monkey_patch_cmake_helper()
    _monkey_patch_autotools_helper()


_do_monkey_patching()



#!/usr/bin/env python3

"""
Settings context manager for recipes
"""

from __future__ import annotations

from multiprocessing import cpu_count
import pathlib
import typing

from qtpy import QtCore
from cruiz.constants import CompilerCacheTypes
from cruiz.exceptions import InconsistentSettingsError

from cruiz.recipe.recipe import Recipe

from .basesettings import (
    BaseSettings,
    BoolSetting,
    StringSetting,
    IntSetting,
    DictSetting,
    WorkflowCwd,
    WorkflowCwdSetting,
    ComparableCommonSettings,
    SettingMeta,
)
from .writermixin import _WriterMixin
from .valueclasses import ScalarValue, DictValue


class RecipeSettings(ComparableCommonSettings):
    """
    Representation of recipe settings
    """

    @classmethod
    def from_recipe(cls, recipe: Recipe) -> RecipeSettings:
        """
        Create an instance using a recipe
        """
        instance = cls()
        instance._recipe = recipe
        return instance

    @classmethod
    def from_uuid(cls, uuid: QtCore.QUuid) -> RecipeSettings:
        """
        Create an instance using a UUID
        """
        instance = cls()
        instance._uuid = uuid
        return instance

    def __init__(self) -> None:
        super().__init__()
        self._uuid: typing.Optional[QtCore.QUuid] = None
        self._recipe: typing.Optional[Recipe] = None

        if self._recipe:
            default_profile = self._recipe.context.default_profile_filename()
        else:
            try:
                context = self.settings_reader.recipe.context  # type: ignore
                default_profile = context.default_profile_filename()
            except AttributeError:
                # this copes with there not being a context assigned
                # a recipe upon first load
                default_profile = None

        self._property_meta = {
            "path": SettingMeta("Path", StringSetting, None, ScalarValue),
            "local_cache_name": SettingMeta(
                "LocalCacheName", StringSetting, None, ScalarValue
            ),
            "profile": SettingMeta(
                "Profile", StringSetting, default_profile, ScalarValue
            ),
            "num_cpu_cores": SettingMeta(
                "NumberOfCores", IntSetting, cpu_count(), ScalarValue
            ),
            "options": SettingMeta("Options", DictSetting, {}, DictValue),
            "attribute_overrides": SettingMeta(
                "Attributes", DictSetting, {}, DictValue
            ),
            "local_workflow_cwd": SettingMeta(
                "LocalWorkflow/Cwd",
                WorkflowCwdSetting,
                WorkflowCwd.RELATIVE_TO_RECIPE,
                ScalarValue,
            ),
            "local_workflow_common_subdir": SettingMeta(
                "LocalWorkflow/CommonSubdirectory", StringSetting, None, ScalarValue
            ),
            "local_workflow_install_folder": SettingMeta(
                "LocalWorkflow/CustomInstallSubdirectory",
                StringSetting,
                None,
                ScalarValue,
            ),
            "local_workflow_imports_folder": SettingMeta(
                "LocalWorkflow/CustomImportsSubdirectory",
                StringSetting,
                None,
                ScalarValue,
            ),
            "local_workflow_source_folder": SettingMeta(
                "LocalWorkflow/CustomSourceSubdirectory",
                StringSetting,
                None,
                ScalarValue,
            ),
            "local_workflow_build_folder": SettingMeta(
                "LocalWorkflow/CustomBuildSubdirectory",
                StringSetting,
                None,
                ScalarValue,
            ),
            "local_workflow_package_folder": SettingMeta(
                "LocalWorkflow/CustomPackageSubdirectory",
                StringSetting,
                None,
                ScalarValue,
            ),
            "local_workflow_test_folder": SettingMeta(
                "LocalWorkflow/CustomTestSubdirectory", StringSetting, None, ScalarValue
            ),
            "cmake_find_debug": SettingMeta(
                "CMake/FindDebug", BoolSetting, False, ScalarValue
            ),
            "cmake_verbose": SettingMeta(
                "CMake/Verbose", BoolSetting, False, ScalarValue
            ),
            "compiler_cache": SettingMeta(
                "CompilerCache", StringSetting, None, ScalarValue
            ),
            "compilercache_autotools_configuration": SettingMeta(
                "CompilerCacheAutoToolsConfig", DictSetting, {}, DictValue
            ),
        }

    @property
    def all_uuids(self) -> typing.List[QtCore.QUuid]:
        """
        Get all UUIDs that have settings.
        """
        settings = BaseSettings.make_settings()
        settings.beginGroup("Recipe")
        uuids = settings.childGroups()
        settings.endGroup()
        uuid_list: typing.List[QtCore.QUuid] = []
        for uuid in uuids:
            uuid_list.append(QtCore.QUuid(uuid))
        return uuid_list

    def matching_uuids(self, recipe_path: pathlib.Path) -> typing.List[QtCore.QUuid]:
        """
        Get UUIDs matching the recipe path provided.
        """
        # TODO: context management around groups would be beneficial here
        settings = BaseSettings.make_settings()
        settings.beginGroup("Recipe")
        uuids = settings.childGroups()
        matches: typing.List[QtCore.QUuid] = []
        for uuid in uuids:
            settings.beginGroup(uuid)
            path = settings.value("Path")
            if not isinstance(path, str):
                raise InconsistentSettingsError()
            if pathlib.Path(path) == recipe_path:
                matches.append(QtCore.QUuid(uuid))
            settings.endGroup()
        settings.endGroup()
        return matches

    @property
    def path(self) -> StringSetting:
        """
        Get the path of the recipe
        """
        return self._get_value_via_meta()

    @path.setter
    def path(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_cache_name(self) -> StringSetting:
        """
        Get the associated local cache name
        """
        return self._get_value_via_meta()

    @local_cache_name.setter
    def local_cache_name(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def profile(self) -> StringSetting:
        """
        Get the profile currently used by the recipe
        """
        return self._get_value_via_meta()

    @profile.setter
    def profile(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def num_cpu_cores(self) -> IntSetting:
        """
        Get the number of CPU cores used by the recipe
        """
        return self._get_value_via_meta()

    @num_cpu_cores.setter
    def num_cpu_cores(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def options(self) -> DictSetting:
        """
        Get the dictionary of recipe options
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "Options"
        options: typing.Dict[str, typing.Any] = {}
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            for i in range(count):
                settings.setArrayIndex(i)
                name = settings.value("Option")
                value = settings.value("Value")
                assert isinstance(name, str)
                options[name] = value
        return DictSetting(options, {})

    def append_options(self, options: DictSetting) -> None:
        """
        Append new option settings

        self.__options set at the end of the function, but in case this is called
        several times, read at the start too
        """
        try:
            existing_options, prop, real_key = self.__options  # type: ignore
        except AttributeError:
            existing_options = {}
            prop = "options"
            real_key = "Options"
        existing_options.update(options)
        # avoids private name mangling
        setattr(  # noqa: B010
            self,
            "__options",
            DictValue(existing_options, prop, real_key, "Option", "Value"),
        )

    @property
    def attribute_overrides(self) -> DictSetting:
        """
        Get the dictionary of attributes
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "Attributes"
        attributes: typing.Dict[str, typing.Any] = {}
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            for i in range(count):
                settings.setArrayIndex(i)
                name = settings.value("Attribute")
                value = settings.value("Value")
                assert isinstance(name, str)
                attributes[name] = value
        return DictSetting(attributes, {})

    def append_attribute(self, attributes: DictSetting) -> None:
        """
        Append new attributes

        self.__attributes set at the end of the function, but in case this is called
        several times, read at the start too
        """
        try:
            existing_attributes, prop, real_key = self.__attributes  # type: ignore
        except AttributeError:
            existing_attributes = {}
            prop = "attribute_overrides"
            real_key = "Attributes"
        existing_attributes.update(attributes)
        # avoids private name mangling
        setattr(  # noqa: B010
            self,
            "__attributes",
            DictValue(existing_attributes, prop, real_key, "Attribute", "Value"),
        )

    @property
    def local_workflow_cwd(self) -> WorkflowCwdSetting:
        """
        Get the local workflow current working directory
        """
        return self._get_value_via_meta()

    @local_workflow_cwd.setter
    def local_workflow_cwd(self, value: WorkflowCwd) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_common_subdir(self) -> StringSetting:
        """
        Get the local workflow common subdirectory to build into
        """
        return self._get_value_via_meta()

    @local_workflow_common_subdir.setter
    def local_workflow_common_subdir(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_install_folder(self) -> StringSetting:
        """
        Get the local workflow Conan install folder
        """
        return self._get_value_via_meta()

    @local_workflow_install_folder.setter
    def local_workflow_install_folder(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_imports_folder(self) -> StringSetting:
        """
        Get the local workflow Conan imports folder
        """
        return self._get_value_via_meta()

    @local_workflow_imports_folder.setter
    def local_workflow_imports_folder(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_source_folder(self) -> StringSetting:
        """
        Get the local workflow Conan source folder
        """
        return self._get_value_via_meta()

    @local_workflow_source_folder.setter
    def local_workflow_source_folder(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_build_folder(self) -> StringSetting:
        """
        Get the local workflow Conan build folder
        """
        return self._get_value_via_meta()

    @local_workflow_build_folder.setter
    def local_workflow_build_folder(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_package_folder(self) -> StringSetting:
        """
        Get the local workflow Conan package folder
        """
        return self._get_value_via_meta()

    @local_workflow_package_folder.setter
    def local_workflow_package_folder(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def local_workflow_test_folder(self) -> StringSetting:
        """
        Get the local workflow Conan test folder
        """
        return self._get_value_via_meta()

    @local_workflow_test_folder.setter
    def local_workflow_test_folder(self, value: str) -> None:
        self._set_value_via_meta(value)

    # TODO: not sure if this is quite how I want to write it
    # also editables are not part of the current UI
    @property
    def editables_count(self) -> IntSetting:
        """
        Get the number of editables in this recipe
        """
        key = "Editables"
        assert self.settings_reader
        with BaseSettings.ReadArray(key, settings=self.settings_reader.settings) as (
            _,
            count,
        ):
            return IntSetting(count, 0)

    @property
    def cmake_find_debug(self) -> BoolSetting:
        """
        Get whether CMake find debug is enabled for this recipe
        """
        return self._get_value_via_meta()

    @cmake_find_debug.setter
    def cmake_find_debug(self, value: bool) -> None:
        self._set_value_via_meta(value)

    @property
    def cmake_verbose(self) -> BoolSetting:
        """
        Get whether CMake verbose mode is enabled for this recipe
        """
        return self._get_value_via_meta()

    @cmake_verbose.setter
    def cmake_verbose(self, value: bool) -> None:
        self._set_value_via_meta(value)

    @property
    def compiler_cache(self) -> StringSetting:
        """
        Get the compiler cache in use with this recipe
        """
        return self._get_value_via_meta()

    @compiler_cache.setter
    def compiler_cache(self, value: bool) -> None:
        self._set_value_via_meta(value)

    @property
    def compilercache_autotools_configuration(self) -> DictSetting:
        """
        Get the dictionary of autotools configuration for this recipe
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "CompilerCacheAutoToolsConfig"
        config: typing.Dict[str, typing.Any] = {}
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            for i in range(count):
                settings.setArrayIndex(i)
                name = settings.value("CompileCache")  # TODO: note the typo
                value = settings.value("ConfigureArgs")
                assert isinstance(name, str)
                config[name] = value
        return DictSetting(config, {})

    def append_compilercache_autotools_configuration(
        self, cache: CompilerCacheTypes, arguments: typing.Optional[str]
    ) -> None:
        """
        Append new autotools compiler cache configurations

        self.__compilercache_autotools_config set at the end of the function,
        but in case this is called several times, read at the start too
        """
        try:
            (
                existing_config,
                prop,
                real_key,
                _,
                _,
            ) = self.__compilercache_autotools_config  # type: ignore
        except AttributeError:
            existing_config = {}
            prop = "compilercache_autotools_configuration"
            real_key = "CompilerCacheAutoToolsConfig"
        existing_config[cache.name] = arguments
        # avoids private name mangling
        setattr(  # noqa: B010
            self,
            "__compilercache_autotools_config",
            DictValue(existing_config, prop, real_key, "CompileCache", "ConfigureArgs"),
        )


class RecipeSettingsReader:
    """
    Context manager for reading recipe settings from disk
    """

    @classmethod
    def from_recipe(cls, recipe: Recipe) -> RecipeSettingsReader:
        """
        Create an instance using a recipe
        """
        return cls(recipe, recipe.uuid)

    @classmethod
    def from_uuid(cls, uuid: QtCore.QUuid) -> RecipeSettingsReader:
        """
        Create an instance using a UUID
        """
        return cls(None, uuid)

    def __init__(self, recipe: typing.Optional[Recipe], uuid: QtCore.QUuid) -> None:
        self.recipe = recipe
        self._uuid = uuid
        self.group = f"Recipe/{uuid.toString(QtCore.QUuid.WithoutBraces)}"
        self.settings = BaseSettings.make_settings()

    def __enter__(self) -> RecipeSettings:
        self.settings.beginGroup(self.group)
        self._settings_object = (
            RecipeSettings.from_recipe(self.recipe)
            if self.recipe
            else RecipeSettings.from_uuid(self._uuid)
        )
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endGroup()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class RecipeSettingsWriter(_WriterMixin):
    """
    Utiiity for writing current recipe settings to disk
    """

    @classmethod
    def from_recipe(cls, recipe: Recipe) -> RecipeSettingsWriter:
        """
        Create an instance using a recipe
        """
        instance = cls()
        instance._reader_for_writer = RecipeSettingsReader.from_recipe(recipe)
        return instance

    @classmethod
    def from_uuid(cls, uuid: QtCore.QUuid) -> RecipeSettingsWriter:
        """
        Create an instance using a UUID
        """
        instance = cls()
        instance._reader_for_writer = RecipeSettingsReader.from_uuid(uuid)
        return instance


class RecipeSettingsDeleter:
    """
    Utility for deleting all recipe settings, associated by a UUID, from disk
    """

    def __init__(self) -> None:
        self._settings = BaseSettings.make_settings()

    def delete(self, uuid: QtCore.QUuid) -> None:
        """
        Delete the recipe settings from the specified UUID
        """
        group = f"Recipe/{uuid.toString(QtCore.QUuid.WithoutBraces)}"
        self._settings.remove(group)



#!/usr/bin/env python3

"""
Mixin for helping to write settings
"""

from enum import Enum

import typing

from qtpy import QtGui

from .basesettings import BaseSettings


class _WriterMixin:
    """
    Mixing class for writing utilities

    self._reader_for_writer defined in the consuming class.
    """

    def sync(self, changes: typing.Any) -> None:
        """
        Synchronise incoming settings changes with those on disk
        """
        assert hasattr(self, "_reader_for_writer")
        keys_to_set = [k for k in changes.__dict__.keys() if k.startswith("__")]
        if not keys_to_set:
            return
        # get current values from settings and fallbacks
        current_values = {}
        assert hasattr(self, "_reader_for_writer")
        with self._reader_for_writer as settings:  # type: ignore
            for key in keys_to_set:
                entry = getattr(changes, key)
                current_values[entry.key] = getattr(settings, entry.property_name)
        # delete those values changes to the fallbacks
        # set others
        with BaseSettings.Group(
            self._reader_for_writer.group  # type: ignore
        ) as settings:  # TODO: should probably re-use the settings instance
            for key in keys_to_set:
                entry = getattr(changes, key)
                # if entry values vs current values are any of the following, the
                # disk setting can be removed
                # - equal
                # - entry is falsey and the fallback is None
                # - entry ie explicitly None
                if (
                    entry.value == current_values[entry.key].fallback
                    or (not entry.value and current_values[entry.key].fallback is None)
                    or entry.value is None
                ):
                    settings.remove(entry.key)
                else:
                    if isinstance(entry.value, Enum):
                        settings.setValue(entry.key, entry.value.value)
                    elif isinstance(entry.value, (str, bool, int, QtGui.QColor)):
                        settings.setValue(entry.key, entry.value)
                    elif isinstance(entry.value, list):
                        # entry.value is the entire list to write
                        # it is checked for empty above
                        assert entry.value
                        with BaseSettings.WriteArray(
                            entry.key, replace=True, settings=settings
                        ) as settings:
                            for i, value in enumerate(entry.value):
                                settings.setArrayIndex(i)
                                settings.setValue(entry.entry_key, value)
                    elif isinstance(entry.value, dict):
                        # entry.value is a delta, include None values for removal
                        current_values[entry.key].value.update(entry.value)
                        dict_entries_to_write = {
                            k: v
                            for k, v in current_values[entry.key].value.items()
                            if v is not None
                        }
                        if dict_entries_to_write:
                            with BaseSettings.WriteArray(
                                entry.key, replace=True, settings=settings
                            ) as settings:
                                for i, (inner_key, value) in enumerate(
                                    dict_entries_to_write.items()
                                ):
                                    settings.setArrayIndex(i)
                                    settings.setValue(entry.entry_key, inner_key)
                                    settings.setValue(entry.entry_value, value)
                        else:
                            settings.remove(entry.key)
                    else:
                        raise RuntimeError(
                            f"Unknown type of data to save as settings: '{entry.key}' "
                            f"is type '{type(entry.value)}'"
                        )
                delattr(changes, key)



#!/usr/bin/env python3

"""
Common parameters
"""

from dataclasses import dataclass, field
import typing


@dataclass
class CommonParameters:
    """
    Common parameters for all commands to run
    """

    worker: typing.Union[
        typing.Callable[[typing.Any, typing.Any], None],
        typing.Callable[[typing.Any, typing.Any, typing.Any], None],
    ]
    added_environment: typing.Dict[str, str] = field(default_factory=dict)
    removed_environment: typing.List[str] = field(default_factory=list)



#!/usr/bin/env python3

"""
Settings context manager for recent Conan remotes
"""

import typing

from .basesettings import BaseSettings, CommonSettings, ListSetting, SettingMeta
from .valueclasses import ListValue
from .writermixin import _WriterMixin


# TODO: CommonSettings not all used
class RecentConanRemotesSettings(CommonSettings):
    """
    Settings representing recent Conan remotes
    """

    def __init__(self) -> None:
        self._property_meta = {
            "urls": SettingMeta("RecentConanRemotes", ListSetting, [], ListValue),
        }

    @property
    def urls(self) -> ListSetting:
        """
        Get all the URLs representing the Conan remotes
        """
        result: typing.List[str] = []
        assert self.settings_reader
        settings = self.settings_reader.settings
        # TODO: remove count on the reader, and replace by
        # num child groups / keys per entry in array
        assert self.settings_reader.count == (
            len(self.settings_reader.settings.childGroups()) / 1
        )
        for i in range(self.settings_reader.count):
            settings.setArrayIndex(i)
            result.append(str(settings.value("URL")))
        return ListSetting(result, [])

    @urls.setter
    def urls(self, urls: typing.List[str]) -> None:
        """
        Set the URLs used for Conan remotes.
        """
        # avoids private name mangling
        setattr(  # noqa: B010
            self, "__new_urls", ListValue(urls, "urls", "RecentConanRemotes", "URL")
        )


class RecentConanRemotesSettingsReader:
    """
    Context manager for reading recent Conan remotes from disk settings
    """

    def __init__(self) -> None:
        self.group = ""
        self.array = "RecentConanRemotes"
        self.settings = BaseSettings.make_settings()
        self.count = 0

    def __enter__(self) -> RecentConanRemotesSettings:
        self.count = self.settings.beginReadArray(self.array)
        self._settings_object = RecentConanRemotesSettings()
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endArray()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class RecentConanRemotesSettingsWriter(_WriterMixin):
    """
    Utility for writing changed recent Conan remote settings to disk
    """

    def __init__(self) -> None:
        self._reader_for_writer = RecentConanRemotesSettingsReader()


class RecentConanRemotesSettingsDeleter:
    """
    Utility for deleting the named recent Conan remote from disk settings
    """

    def __init__(self) -> None:
        self._reader_for_deleter = RecentConanRemotesSettingsReader()

    def delete(self, url: str) -> None:
        """
        Delete a specific recent Conan remote URL from disk settings
        """
        with self._reader_for_deleter as settings:
            current_list = settings.urls.resolve()
        current_list = [x for x in current_list if x != url]
        BaseSettings.make_settings().remove(self._reader_for_deleter.array)
        if current_list:
            with BaseSettings.WriteArray(self._reader_for_deleter.array) as settings:
                for i, path in enumerate(current_list):
                    settings.setArrayIndex(i)
                    settings.setValue("URL", path)
